"""Flsqls module."""


from pineboolib.core import decorators


from pineboolib.application.metadata import pntablemetadata
from pineboolib import logging

from pineboolib.fllegacy import flutil
from pineboolib.interfaces import isqldriver


from typing import Optional, Union, List, Any, Dict, TYPE_CHECKING

if TYPE_CHECKING:
    from sqlalchemy.engine import (  # type: ignore [import] # noqa: F401, F821
        base,  # noqa: F401
        result,  # noqa: F401
    )  # noqa: F401 # pragma: no cover


LOGGER = logging.get_logger(__name__)


class FLPYMSSQL(isqldriver.ISqlDriver):
    """FLPYMSSQL class."""

    def __init__(self):
        """Inicialize."""
        super().__init__()
        self.version_ = "0.9"
        self.name_ = "FLPYMSSQL"
        self.error_list = []
        self.alias_ = "SQL Server (PYMSSQL)"
        self.default_port = 1433
        self._like_true = "1"
        self._like_false = "0"
        self._safe_load = {"pymssql": "pymssql", "sqlalchemy": "sqlAlchemy"}
        self._database_not_found_keywords = ["20018"]
        self._text_like = ""
        self._sqlalchemy_name = "mssql+pymssql"
        self._create_isolation = False

    def getAlternativeConn(self, name: str, host: str, port: int, usern: str, passw_: str) -> Any:
        """Return connection."""
        self._queqe_params["connect_args"] = {"autocommit": True}
        conn_ = self.getConn("master", host, port, usern, passw_)
        del self._queqe_params["connect_args"]
        # conn_.execute("set transaction isolation level read uncommitted;")
        return conn_

    # def session(self) -> "isession.PinebooSession":
    #    """Create a sqlAlchemy session."""
    #    while True:
    #        session_class = sessionmaker(bind=self.connection(), autoflush=False, autocommit=True)

    #        new_session = session_class()
    #        if new_session.connection().connection is not None:
    #            break
    #        else:
    #            LOGGER.warning("Conexión invalida capturada.Solicitando nueva")

    #    setattr(new_session, "_conn_name", self.db_._name)
    #    session_key = utils_base.session_id(self.db_._name, True)
    #    self.db_._conn_manager._thread_sessions[session_key] = new_session
    #    return new_session

    def existsTable(self, table_name: str) -> bool:
        """Return if exists a table specified by name."""

        sql = (
            "SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES WHERE "
            + "TABLE_NAME = N'%s' AND TABLE_CATALOG = '%s'" % (table_name, self._dbname)
        )
        cur = self.execute_query(sql)
        return True if cur and cur.fetchone() else False

    def nextSerialVal(self, table_name: str, field_name: str) -> int:
        """Return next serial value."""

        if self.is_open():
            cur = self.execute_query("SELECT NEXT VALUE FOR %s_%s_seq" % (table_name, field_name))

            if cur and cur.returns_rows:  # type: ignore [attr-defined]
                return cur.fetchone()[0]  # type: ignore [index] # noqa: F821

            LOGGER.warning("not exec sequence")

        return 0

    def releaseSavePoint(self, num: int) -> bool:
        """Set release savepoint."""

        return True

    def setType(self, type_: str, leng: int = 0) -> str:
        """Return type definition."""
        type_ = type_.lower()
        res_ = ""
        if type_ in ("int", "serial"):
            res_ = "INT"
        elif type_ == "uint":
            res_ = "BIGINT"
        elif type_ in ("bool", "unlock"):
            res_ = "BIT"
        elif type_ == "double":
            res_ = "DECIMAL"
        elif type_ == "time":
            res_ = "TIME"
        elif type_ == "date":
            res_ = "DATE"
        elif type_ in ("pixmap", "stringlist"):
            res_ = "TEXT"
        elif type_ == "string":
            res_ = "VARCHAR"
        elif type_ == "bytearray":
            res_ = "NVARCHAR"
        elif type_ == "timestamp":
            res_ = "DATETIME2"
        elif type_ == "json":
            res_ = "NVARCHAR"
        else:
            LOGGER.warning("seType: unknown type %s", type_)
            leng = 0

        return "%s(%s)" % (res_, leng) if leng else res_

    def sqlCreateTable(
        self, tmd: "pntablemetadata.PNTableMetaData", create_index: bool = True
    ) -> Optional[str]:
        """Return a create table query."""

        if tmd.isQuery():
            return self.sqlCreateView(tmd)

        util = flutil.FLUtil()

        primary_key = ""
        sql = "CREATE TABLE %s (" % tmd.name()
        seq = None

        field_list = tmd.fieldList()

        unlocks = 0
        for number, field in enumerate(field_list):
            sql += field.name()
            type_ = field.type()
            if type_ == "serial":
                seq = "%s_%s_seq" % (tmd.name(), field.name())
                if self.is_open() and create_index:
                    try:
                        self.execute_query("CREATE SEQUENCE %s START WITH 1 INCREMENT BY 1" % seq)
                    except Exception as error:
                        LOGGER.error("%s::sqlCreateTable:%s", __name__, str(error))

                    sql += " INT"

            elif type_ == "double":
                sql += " DECIMAL(%s,%s)" % (
                    int(field.partInteger()) + int(field.partDecimal()),
                    int(field.partDecimal()),
                )

            else:
                if type_ == "unlock":
                    unlocks += 1

                    if unlocks > 1:
                        LOGGER.warning("FLManager : No se ha podido crear la tabla %s ", tmd.name())
                        LOGGER.warning(
                            "FLManager : Hay mas de un campo tipo unlock. Solo puede haber uno."
                        )
                        return None

                sql += " %s" % self.setType(type_, field.length())

            if field.isPrimaryKey():
                if not primary_key:
                    sql = sql + " PRIMARY KEY"
                    primary_key = field.name()
                else:
                    LOGGER.warning(
                        util.translate(
                            "application",
                            "FLManager : Tabla-> %s ." % tmd.name()
                            + "Se ha intentado poner una segunda clave primaria para el campo %s ,pero el campo %s ya es clave primaria."
                            % (primary_key, field.name())
                            + "Sólo puede existir una clave primaria en FLTableMetaData, use FLCompoundKey para crear claves compuestas.",
                        )
                    )
                    raise Exception(
                        "A primary key (%s) has been defined before the field %s.%s -> %s"
                        % (primary_key, tmd.name(), field.name(), sql)
                    )
            else:
                sql += " UNIQUE" if field.isUnique() else ""
                sql += " NULL" if field.allowNull() else " NOT NULL"

            if number != len(field_list) - 1:
                sql += ","

        sql += ")"

        return sql

    def decodeSqlType(self, type_: Union[int, str]) -> str:
        """Return the specific field type."""
        ret = str(type_).lower()

        if type_ == "bit":
            ret = "bool"
        elif type_ == "bigint":
            ret = "uint"
        elif type_ == "decimal":
            ret = "double"
        elif type_ == "date":
            ret = "date"
        elif type_ == "time":
            ret = "time"
        elif type_ == "varchar":
            ret = "string"
        elif type_ == "nvarchar":
            ret = "bytearray"
        elif type_ == "text":
            ret = "stringlist"
        elif type_ == "datetime2":
            ret = "timestamp"
        elif type_ == "json":
            ret = "json"

        return ret

    def tables(self, type_name: str = "", table_name: str = "") -> List[str]:
        """Return a tables list specified by type."""
        table_list: List[str] = []
        result_list: List[Any] = []
        if self.is_open():
            where: List[str] = []
            if type_name in ("Tables", ""):
                where.append("xtype ='U'")
            if type_name in ("Views", ""):
                where.append("xtype ='V'")
            if type_name in ("SystemTables", ""):
                where.append("xtype ='S'")

            if where:
                and_name = " AND name ='%s'" % (table_name) if table_name else ""

                cursor = self.execute_query(
                    "SELECT name FROM SYSOBJECTS where %s%s ORDER BY name ASC"
                    % (" OR ".join(where), and_name)
                )
                result_list += cursor.fetchall() if cursor else []

        table_list = [item[0] for item in result_list]

        return table_list

    def declareCursor(
        self, curname: str, fields: str, table: str, where: str, conn_db: "base.Connection"
    ):
        """Set a refresh query for database."""

        if not self.is_open():
            raise Exception("declareCursor: Database not open")

        sql = "DECLARE %s CURSOR STATIC FOR SELECT %s FROM %s WHERE %s " % (
            curname,
            fields,
            table,
            where,
        )
        try:
            conn_db.execute(sql)
            conn_db.execute("OPEN %s" % curname)
        except Exception as error:
            LOGGER.error("refreshQuery: %s", error)
            LOGGER.info("SQL: %s", sql)
            LOGGER.trace("Detalle:", stack_info=True)

    def deleteCursor(self, cursor_name: str, cursor: Any) -> None:
        """Delete cursor."""

        if not self.is_open():
            raise Exception("deleteCursor: Database not open")

        try:
            sql_exists = "SELECT CURSOR_STATUS('global','%s')" % cursor_name
            cursor.execute(sql_exists)

            if cursor.fetchone()[0] < 1:
                return

            cursor.execute("CLOSE %s" % cursor_name)
        except Exception as exception:
            LOGGER.error("finRow: %s", exception)
            LOGGER.warning("Detalle:", stack_info=True)

    # def fix_query(self, query: str) -> str:
    #    """Fix string."""
    #    # ret_ = query.replace(";", "")
    #    return query

    @decorators.not_implemented_warn
    def alterTable(self, new_metadata: "pntablemetadata.PNTableMetaData") -> bool:
        """Modify a table structure."""

        return False

    def recordInfo2(self, tablename: str) -> Dict[str, List[Any]]:
        """Return info from a database table."""
        info = {}
        sql = (
            "SELECT COLUMN_NAME, DATA_TYPE, IS_NULLABLE, COLUMN_DEFAULT, NUMERIC_PRECISION_RADIX,"
            + " CHARACTER_MAXIMUM_LENGTH FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME = '%s'"
            % tablename.lower()
        )

        data = self.execute_query(sql)
        res = data.fetchall() if data else []
        for columns in res:
            field_size = int(columns[5]) if columns[5] else 0
            # field_precision = columns[4] or 0
            field_name = columns[0]
            field_type = self.decodeSqlType(columns[1])
            field_allow_null = columns[2] == "YES"
            field_default_value = columns[3]

            info[field_name] = [
                field_name,
                field_type,
                not field_allow_null,
                field_size,
                None,
                field_default_value,
                None,  # field_pk
            ]

        return info

    def vacuum(self) -> None:
        """Vacuum tables."""

        return

    def sqlLength(self, field_name: str, size: int) -> str:
        """Return length formated."""

        return "LEN(%s)=%s" % (field_name, size)



"""Test_sys module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing


class TestSys(unittest.TestCase):
    """TestSys Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_afterCommit_flFiles(self) -> None:
        """Test aftercommit_flfiles function."""
        from pineboolib.application.database import pnsqlcursor

        cur_area = pnsqlcursor.PNSqlCursor("flareas")
        cur_area.setModeAccess(cur_area.Insert)
        cur_area.refreshBuffer()
        cur_area.setValueBuffer("idarea", "A")
        cur_area.setValueBuffer("descripcion", "area A")
        self.assertTrue(cur_area.commitBuffer())
        _cur_modulo = pnsqlcursor.PNSqlCursor("flmodules")
        _cur_modulo.setModeAccess(_cur_modulo.Insert)
        _cur_modulo.refreshBuffer()
        _cur_modulo.setValueBuffer("idarea", "A")
        _cur_modulo.setValueBuffer("idmodulo", "TM")
        _cur_modulo.setValueBuffer("descripcion", "modulo TM")
        self.assertTrue(_cur_modulo.commitBuffer())
        _cur = pnsqlcursor.PNSqlCursor("flfiles")
        _cur.setModeAccess(_cur.Insert)
        _cur.refreshBuffer()
        _cur.setValueBuffer("nombre", "prueba")
        _cur.setValueBuffer("idmodulo", "TM")
        _cur.setValueBuffer("contenido", "Pablito clavó un clavito!")
        self.assertTrue(_cur.commitBuffer())

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



# -*- coding: utf-8 -*-
"""
AQOds package.

Generate .ods files (Opendocument Spreadsheet)
"""
from typing import Union, List, Any, Tuple, Optional, TYPE_CHECKING
from pineboolib.core.utils import utils_base

if not utils_base.is_library():
    try:
        from odf import table, style  # type: ignore
    except ImportError:
        pass

from pineboolib import logging

if TYPE_CHECKING:
    from odf.opendocument import OpenDocumentSpreadsheet  # type: ignore # pragma: no cover
    from odf import table, style  # type: ignore
    import odf  # type: ignore


LOGGER = logging.get_logger(__name__)

"""
Generador de ficheros ODS
"""


class OdsStyleFlags(object):
    """OdsStyleFlags."""

    ODS_NONE = 0
    ODS_BORDER_BOTTOM = 10
    ODS_BORDER_LEFT = 11
    ODS_BORDER_RIGHT = 12
    ODS_BORDER_TOP = 13
    ODS_ALIGN_LEFT = 14
    ODS_ALIGN_CENTER = 15
    ODS_ALIGN_RIGHT = 16
    ODS_TEXT_BOLD = 17
    ODS_TEXT_ITALIC = 18
    ODS_TEXT_UNDERLINE = 19


class AQOdsGenerator(object):
    """AQOdsGenerator Class."""

    doc_: Any = None
    """
    Constructor
    """

    def __init__(self):
        """
        Initialize the class by checking dependencies.
        """

        from pineboolib.application.utils.check_dependencies import check_dependencies

        check_dependencies({"odf": "odfpy"})

    def generateOds(self, file_name: str) -> None:
        """
        Generate the ODS file.

        @param file_name. File name to generate.
        """
        if self.doc_ is None:
            raise Exception("Document not set, cannot generate")
        file_name = file_name.replace(".ods", "")
        self.doc_.save(file_name, True)

    def set_doc_(self, document) -> None:
        """
        Assign the contents of the file.

        @param document. Data to add to the file
        """

        self.doc_ = document


class AQOdsSpreadSheet(object):
    """
    AQOdsSpreadSheet Class.

    Generate ODS document.
    """

    spread_sheet: "OpenDocumentSpreadsheet"
    generator_: "AQOdsGenerator"

    def __init__(self, generator: "AQOdsGenerator") -> None:
        """
        Initialize the file generator.

        @param generator. File generator.
        """

        self.generator_ = generator
        from odf.opendocument import OpenDocumentSpreadsheet  # Fuera rompe deploy

        self.spread_sheet = OpenDocumentSpreadsheet()
        self.generator_.set_doc_(self.spread_sheet)

    def close(self) -> None:
        """
        Close the document.
        """

        pass


class AQOdsImage(object):
    """AQOdsImage Class."""

    name_: str
    width_: float
    height_: float
    pos_x: int
    pos_y: int
    link_: str

    def __init__(self, name: str, width: float, height: float, x: int, y: int, link: str) -> None:
        """
        Initialize the Image.

        @param name. Image Name
        @param width. Width.
        @param height. Tall.
        @param x. X position.
        @param me. Position and.
        @param link. Image link.
        """

        self.name_ = name
        self.width_ = width
        self.height_ = height
        self.pos_x = x
        self.pos_y = y
        self.link_ = link


class AQOdsSheet(object):
    """
    AQOdsSheet Class.

    Sheet inside the document.
    """

    num_rows_: int
    spread_sheet_parent_: "OpenDocumentSpreadsheet"
    sheet_: "table.Table"

    def __init__(self, spread_sheet: AQOdsSpreadSheet, sheet_name: str) -> None:
        """
        Initialize the sheet.

        @param spread_sheet. Spreadsheet.
        @param sheet_name. Name of the sheet.
        """
        from odf import table

        self.spread_sheet_parent_ = spread_sheet.spread_sheet
        self.num_rows_ = 0

        self.sheet_ = table.Table(name=sheet_name)

    def rowsCount(self) -> int:
        """
        Return number of lines.

        @return Number of lines.
        """

        return self.num_rows_

    def close(self) -> None:
        """Close the sheet."""

        self.spread_sheet_parent_.spreadsheet.addElement(self.sheet_)


class AQOdsRow(object):
    """AQOdsRow."""

    sheet_: AQOdsSheet
    row_: "table.TableRow"
    cells_list_: List["table.TableCell"]
    style_cell_text_: Optional[str]
    fix_precision_: Optional[int]
    row_color_: Optional[str]
    property_cell_: List[Any]

    def __init__(self, sheet: AQOdsSheet) -> None:
        """
        Initialize a line inside a sheet.

        @param sheet. Parent Document Sheet.
        """
        from odf import table  # type: ignore

        self.sheet_ = sheet

        self.row_ = table.TableRow()
        self.cells_list_ = []
        self.fix_precision_ = None
        self.row_color_ = None
        self.property_cell_ = []
        self.style_cell_text_ = None

    def addBgColor(self, color) -> None:
        """
        Specify the background color of the line.

        @param color. It is specified in hex (color) format [2:]
        """

        self.row_color_ = color

    def opIn(self, opt: Union[float, str, "odf.element.Element", List, "AQOdsImage"]):
        """
        Add options to the line.

        @param opt. Line options, each cell ends with the value assignment
        """

        from odf.text import P, Span  # type: ignore
        from odf.draw import Frame, Image  # type: ignore
        from odf import element as odf_elem  # type: ignore

        if isinstance(opt, float):
            if self.fix_precision_ is not None:
                opt = "%s" % round(opt, self.fix_precision_)
            else:
                opt = "%s" % opt
        if isinstance(opt, str):  # Último paso
            cell, style = self.__newCell__()

            if self.style_cell_text_:
                text_elem = P(text="")
                txt_ = Span(stylename=self.style_cell_text_, text=opt)
                text_elem.addElement(txt_)
            else:
                text_elem = P(text=opt)

            self.sheet_.spread_sheet_parent_.automaticstyles.addElement(style)
            cell.addElement(text_elem)
            self.cells_list_.append(cell)
            self.fix_precision_ = None
            self.style_cell_text_ = None

        else:
            if isinstance(opt, list):  # Si es lista , Insertamos todos los parámetros uno a uno
                for item in opt:
                    self.opIn(item)

            elif isinstance(opt, AQOdsImage):
                href = self.sheet_.spread_sheet_parent_.addPictureFromFile(opt.link_)
                cell, style = self.__newCell__()

                # p = P()
                frame = Frame(
                    width="%spt" % opt.width_,
                    height="%spt" % opt.height_,
                    x="%spt" % opt.pos_x,
                    y="%spt" % opt.pos_y,
                )
                frame.addElement(Image(href=href))
                # p.addElement(frame)
                cell.addElement(frame)
                self.cells_list_.append(cell)
                # self.coveredCell()
                # self.opIn(href)
                # print("FIXME:: Vacio", href)
            elif isinstance(opt, odf_elem.Element):
                if opt.tagName in ("style:paragraph-properties", "style:table-cell-properties"):
                    import copy

                    prop = copy.copy(opt)
                    self.property_cell_.append(prop)
                elif opt.tagName == "style:style":
                    self.sheet_.spread_sheet_parent_.automaticstyles.addElement(opt)
                    self.style_cell_text_ = opt
                else:
                    LOGGER.warning("%s:Parámetro desconocido %s", __name__, opt.tagName)

    def __newCell__(self) -> Tuple["table.TableCell", "style.Style"]:
        """
        Return new cell This is created by assigning the value to the previous one.

        @return Tuple(TableCell, Style)
        """

        from odf import table, style  # type: ignore

        style_cell = style.Style(
            name="stylecell_%s_%s" % (len(self.cells_list_), self.sheet_.rowsCount()),
            family="table-cell",
        )
        if self.row_color_:  # Guardo color si hay
            style_cell.addElement(
                style.TableCellProperties(backgroundcolor="#%s" % self.row_color_)
            )

        for prop in self.property_cell_:  # Guardo prop cell si hay
            style_cell.addElement(prop)

        self.property_cell_ = []
        return table.TableCell(valuetype="string", stylename=style_cell), style_cell

    def close(self) -> None:
        """
        Close the line.
        """

        for cell in self.cells_list_:  # Meto las celdas en la linea
            self.row_.addElement(cell)

        self.sheet_.num_rows_ += 1  # Especifico cunatas lineas tiene ya la hoja
        self.sheet_.sheet_.addElement(self.row_)  # Meto la nueva linea en la hoja

    def coveredCell(self) -> None:
        """
        Field is filled with empty data.
        """

        self.opIn(" ")

    def setFixedPrecision(self, num_: Optional[int]) -> None:
        """
        Specify the precision of number decimals.

        @param n. Decimal precision.
        """
        if num_ is not None:
            self.fix_precision_ = num_


def aq_ods_color(*args) -> str:
    """
    Return a color from a hexadecimal value.

    @param color. Hexadecimal value.
    """
    if len(args) == 1:
        return hex(args[0])[2:]
    else:
        return "%02x%02x%02x" % (args[0], args[1], args[2])


class AQOdsStyle(object):
    """AQOdsStyle class."""

    def __init__(self, style: Optional[int] = 0):
        """Initialize."""

        if style is not None:
            ods = OdsStyleFlags
            if style == ods.ODS_NONE:
                return
            elif style == ods.ODS_BORDER_BOTTOM:
                self = self.borderBottom()
            elif style == ods.ODS_BORDER_LEFT:
                self = self.borderLeft()
            elif style == ods.ODS_BORDER_RIGHT:
                self = self.borderRight()
            elif style == ods.ODS_BORDER_TOP:
                self = self.borderTop()
            elif style == ods.ODS_ALIGN_LEFT:
                self = self.alignLeft()
            elif style == ods.ODS_ALIGN_CENTER:
                self = self.alignCenter()
            elif style == ods.ODS_ALIGN_RIGHT:
                self = self.alignRight()
            elif style == ods.ODS_TEXT_BOLD:
                self = self.textBold()
            elif style == ods.ODS_TEXT_ITALIC:
                self = self.textItalic()
            elif style == ods.ODS_TEXT_UNDERLINE:
                self = self.textUnderline()
            else:
                raise ValueError("Unknown flag!")

    def alignCenter(self) -> "style.ParagraphProperties":
        """
        Align a cell to the center property.

        @return Style Property.
        """
        from odf import style  # type: ignore

        return style.ParagraphProperties(textalign="center")

    def alignRight(self) -> "style.ParagraphProperties":
        """
        Align a cell to the right property.

        @return Style Property.
        """
        from odf import style  # type: ignore

        return style.ParagraphProperties(textalign="right")

    def alignLeft(self) -> "style.ParagraphProperties":
        """
        Align a cell to the left property.

        @return Style Property.
        """
        from odf import style  # type: ignore

        return style.ParagraphProperties(textalign="left")

    def textBold(self) -> "style.Style":
        """
        Return text bold property.

        @return Style Property.
        """
        from odf import style  # type: ignore

        bold_style = style.Style(name="Bold", family="text")
        bold_style.addElement(style.TextProperties(fontweight="bold"))
        return bold_style

    def textUnderline(self) -> "style.Style":
        """
        Return text bold property.

        @return Style Property.
        """
        from odf import style  # type: ignore

        bold_style = style.Style(name="Underline", family="text")
        bold_style.addElement(style.TextProperties(fontstyle="underline"))
        return bold_style

    def textItalic(self) -> "style.Style":
        """
        Return italic text.

        @return Style Property.
        """
        from odf import style  # type: ignore

        italic_style = style.Style(name="Italic", family="text")
        italic_style.addElement(style.TextProperties(fontstyle="italic"))
        return italic_style

    def borderBottom(self) -> "style.TableCellProperties":
        """
        Return the property of the lower edge of a cell.

        @return Table cell property.
        """

        from odf import style  # type: ignore

        return style.TableCellProperties(borderbottom="1pt solid #000000")

    def borderLeft(self) -> "style.TableCellProperties":
        """
        Return the property of the left edge of a cell.

        @return Table cell property.
        """

        from odf import style  # type: ignore

        return style.TableCellProperties(borderleft="1pt solid #000000")

    def borderRight(self) -> "style.TableCellProperties":
        """
        Return the property of the right edge of a cell.

        @return Table cell property.
        """

        from odf import style  # type: ignore

        return style.TableCellProperties(borderright="1pt solid #000000")

    def borderTop(self) -> "style.TableCellProperties":
        """
        Return the property of the upper edge of a cell.

        @return Table cell property.
        """

        from odf import style  # type: ignore

        return style.TableCellProperties(bordertop="1pt solid #000000")

    Align_center = property(alignCenter, None)
    Align_right = property(alignRight, None)
    Align_left = property(alignLeft, None)
    Text_bold = property(textBold, None)
    Text_italic = property(textItalic, None)
    Text_underline = property(textUnderline, None)
    Border_bottom = property(borderBottom, None)
    Border_top = property(borderTop, None)
    Border_right = property(borderRight, None)
    Border_left = property(borderLeft, None)



# -*- coding: utf-8 -*-
# Translated with pineboolib v0.71.18
"""Flsetting_model module."""

import sqlalchemy  # type: ignore [import] # noqa: F821

from pineboolib.application.database.orm import basemodel


class Flsettings(basemodel.BaseModel):  # type: ignore [misc] # noqa: F821
    """Flsetting class."""

    __tablename__ = "flsettings"

    # --- Metadata --->
    legacy_metadata = {
        "name": "flsettings",
        "alias": "Configuración global",
        "fields": [
            {
                "name": "flkey",
                "alias": "Clave",
                "pk": True,
                "type": "string",
                "length": 30,
                "null": False,
            },
            {"name": "valor", "alias": "Valor", "type": "stringlist", "visiblegrid": False},
        ],
    }

    # <--- Metadata ---

    # --- Fields --->

    flkey = sqlalchemy.Column("flkey", sqlalchemy.String(30), primary_key=True)
    valor = sqlalchemy.Column("valor", sqlalchemy.String)


# <--- Fields ---



"""Systype module."""

import traceback
import os
import os.path
import sys
import re


from PyQt6 import QtCore, QtWidgets, QtGui, QtXml  # type: ignore[import]


from pineboolib.core.utils import utils_base, logging

from pineboolib.core import decorators, settings, translate

from pineboolib import application
from pineboolib.application import types, process

from pineboolib.application.database import pnsqlcursor, pnsqlquery, utils as utils_db


from pineboolib.application.packager import pnunpacker
from pineboolib.application.qsatypes import sysbasetype


from pineboolib.fllegacy.aqsobjects import aqsql, aqs

from pineboolib.fllegacy import flutil
from pineboolib.fllegacy import flvar
from pineboolib.fllegacy import flfielddb
from pineboolib.fllegacy import fltabledb

from pineboolib.q3widgets import (
    qdialog,
    qlabel,
    qtextedit,
    qvboxlayout,
    qhboxlayout,
    qpushbutton,
    filedialog,
    dialog,
    qbytearray,
    messagebox,
)

from typing import cast, Optional, List, Any, Dict, Callable, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.interfaces import iconnection, isqlcursor  # pragma: no cover
    from pineboolib.fllegacy import flformrecorddb, flformdb  # noqa: F401 # pragma: no cover


LOGGER = logging.get_logger(__name__)


class AQTimer(QtCore.QTimer):
    """AQTimer class."""

    pass


class AQGlobalFunctionsClass(QtCore.QObject):
    """AQSGlobalFunction class."""

    functions_: Dict[str, Callable]
    mappers_: "QtCore.QSignalMapper"

    def __init__(self):
        """Initialize."""

        super().__init__()
        self.functions_ = {}
        self.mappers_ = QtCore.QSignalMapper()

    def set(self, function_name: str, global_function: Callable) -> None:
        """Set a new global function."""
        self.functions_[function_name] = global_function

    def get(self, function_name: str) -> Callable:
        """Return a global function specified by name."""

        return self.functions_[function_name]

    def exec_(self, function_name: str) -> None:
        """Execute a function specified by name."""

        fun = self.functions_[function_name]
        if fun is not None:
            fun()

    def mapConnect(self, obj: "QtWidgets.QWidget", signal: str, function_name: str) -> None:
        """Add conection to map."""

        self.mappers_.mappedString.connect(self.exec_)  # type: ignore
        sg_name = re.sub(r" *\(.*\)", "", signal)

        signal_ = getattr(obj, sg_name, None)
        if signal_ is not None:
            signal_.connect(self.mappers_.map)
            self.mappers_.setMapping(obj, function_name)


class SysType(sysbasetype.SysBaseType):
    """SysType class."""

    time_user_ = QtCore.QDateTime.currentDateTime()
    AQTimer = AQTimer
    AQGlobalFunctions = AQGlobalFunctionsClass()

    @classmethod
    def translate(cls, *args) -> str:
        """Translate a text."""

        group, text = (args[0], args[1]) if len(args) == 2 else ("scripts", args[0])

        if text == "MetaData":
            group, text = text, group

        return translate.translate(group, text.replace(" % ", " %% "))

    def printTextEdit(self, editor: "QtWidgets.QTextEdit"):
        """Print text from a textEdit."""

        application.PROJECT.aq_app.printTextEdit(editor)

    def diskCacheAbsDirPath(self) -> str:
        """Return Absolute disk cache path."""

        return os.path.abspath(application.PROJECT.tmpdir)

    def dialogGetFileImage(self) -> Optional[str]:
        """Show a file dialog and return a file name."""

        return application.PROJECT.aq_app.dialogGetFileImage()

    def toXmlReportData(self, qry: "pnsqlquery.PNSqlQuery") -> "QtXml.QDomDocument":
        """Return xml from a query."""

        return application.PROJECT.aq_app.toXmlReportData(qry)

    def showDocPage(self, url_: str) -> None:
        """Show externa file."""

        return application.PROJECT.aq_app.showDocPage(url_)

    def toPixmap(self, value_: str) -> "QtGui.QPixmap":
        """Create a QPixmap from a text."""

        return application.PROJECT.aq_app.toPixmap(value_)

    def setMultiLang(self, enable_: bool, lang_id_: str) -> None:
        """
        Change multilang status.

        @param enable, Boolean con el nuevo estado
        @param langid, Identificador del leguaje a activar
        """

        return application.PROJECT.aq_app.setMultiLang(enable_, lang_id_)

    def fromPixmap(self, pix_: "QtGui.QPixmap") -> str:
        """Return a text from a QPixmap."""

        return application.PROJECT.aq_app.fromPixmap(pix_)

    def popupWarn(self, msg_warn: str, script_calls: List[Any] = []) -> None:
        """Show a warning popup."""

        application.PROJECT.aq_app.popupWarn(msg_warn, script_calls)

    def openMasterForm(self, action_name_: str) -> None:
        """Open default form from a action."""

        if action_name_ in application.PROJECT.actions.keys():
            application.PROJECT.actions[action_name_].openDefaultForm()

    def scalePixmap(
        self, pix_: "QtGui.QPixmap", width_: int, height_: int, mode_: "QtCore.Qt.AspectRatioMode"
    ) -> "QtGui.QImage":
        """Return QImage scaled from a QPixmap."""

        return application.PROJECT.aq_app.scalePixmap(pix_, width_, height_, mode_)

    @classmethod
    def transactionLevel(cls) -> int:
        """Return transaction level."""

        return application.PROJECT.conn_manager.useConn("default").transactionLevel()

    @classmethod
    def installACL(cls, idacl) -> None:
        """Install a acl."""
        from pineboolib.application.acls import pnaccesscontrollists

        acl_ = pnaccesscontrollists.PNAccessControlLists()

        if acl_:
            acl_.install_acl(idacl)

    @classmethod
    def updateAreas(cls) -> None:
        """Update areas in mdi."""
        func_ = getattr(application.PROJECT.main_window, "initToolBox", None)
        if func_ is not None:
            func_()

    @classmethod
    def reinit(cls) -> None:
        """Call reinit script."""

        while application.PROJECT.aq_app._inicializing:
            application.PROJECT.app.processEvents()  # type: ignore [misc]

        application.PROJECT.aq_app.reinit()

    @classmethod
    def modMainWidget(cls, id_module_: str) -> Optional["QtWidgets.QWidget"]:
        """Set module MainWinget."""

        return application.PROJECT.aq_app.modMainWidget(id_module_)

    @classmethod
    def setCaptionMainWidget(cls, title: str) -> None:
        """Set caption in the main widget."""

        application.PROJECT.aq_app.setCaptionMainWidget(title)

    @staticmethod
    def execQSA(qsa_file=None, args=None) -> Any:
        """Execute a QS file."""

        try:
            file_ = open(qsa_file, "r")
            data = file_.read()
            file_.close()
            fun = types.function("exec_qsa", data)
            return fun(args)
        except Exception:
            error = traceback.format_exc()
            LOGGER.warning(error)

        return None

    @staticmethod
    def dumpDatabase() -> None:
        """Launch dump database."""
        aq_dumper = AbanQDbDumper()
        aq_dumper.init()

    @staticmethod
    def terminateChecksLocks(cursor: Optional["isqlcursor.ISqlCursor"] = None) -> None:
        """Set check risk locks to False in a cursor."""
        if cursor is not None:
            cursor.checkRisksLocks(True)

    @classmethod
    def mvProjectXml(cls) -> "QtXml.QDomDocument":
        """Extract a module defition to a QDomDocument."""

        doc_ret_ = QtXml.QDomDocument()
        str_xml_ = utils_db.sql_select("flupdates", "modulesdef", "actual")
        if not str_xml_:
            return doc_ret_
        doc = QtXml.QDomDocument()
        if not doc.setContent(str_xml_):
            return doc_ret_
        str_xml_ = ""
        nodes = doc.childNodes()

        for number in range(len(nodes)):
            it_ = nodes.item(number)
            if it_.isComment():
                data = it_.toComment().data()
                if not data == "" and data.startswith("<mvproject "):
                    str_xml_ = data
                    break

        if str_xml_ == "":
            return doc_ret_
        doc_ret_.setContent(str_xml_)
        return doc_ret_

    @classmethod
    def mvProjectModules(cls) -> "types.Array":
        """Return modules defitions Dict."""
        ret = types.Array()
        doc = cls.mvProjectXml()
        mods = doc.elementsByTagName("module")
        for number in range(len(mods)):
            it_ = mods.item(number).toElement()
            mod = {"name": (it_.attribute("name")), "version": (it_.attribute("version"))}
            if len(mod["name"]) == 0:
                continue
            ret[mod["name"]] = mod

        return ret

    @classmethod
    def mvProjectExtensions(cls) -> "types.Array":
        """Return project extensions Dict."""

        ret = types.Array()
        doc = cls.mvProjectXml()
        exts = doc.elementsByTagName("extension")

        for number in range(len(exts)):
            it_ = exts.item(number).toElement()
            ext = {"name": (it_.attribute("name")), "version": (it_.attribute("version"))}
            if len(ext["name"]) == 0:
                continue
            ret[ext["name"]] = ext

        return ret

    @classmethod
    def calculateShaGlobal(cls) -> str:
        """Return sha global value."""

        value = ""
        qry = pnsqlquery.PNSqlQuery()
        qry.setSelect("sha")
        qry.setFrom("flfiles")
        if qry.exec_() and qry.first():
            value = utils_base.sha1(str(qry.value(0)))
            while qry.next():
                value = utils_base.sha1(value + str(qry.value(0)))
        else:
            value = utils_base.sha1("")

        return value

    def registerUpdate(self, input_: str = "") -> None:
        """Install a package."""

        if not input_:
            return
        unpacker = pnunpacker.PNUnpacker(input_)
        errors = unpacker.errorMessages()
        if len(errors) != 0:
            msg = self.translate("Hubo los siguientes errores al intentar cargar los módulos:")
            msg += "\n"
            for number in range(len(errors)):
                msg += utils_base.ustr(errors[number], "\n")

            self.errorMsgBox(msg)
            return

        unpacker.jump()
        unpacker.jump()
        unpacker.jump()
        now = str(types.Date())
        file_ = types.File(input_)
        file_name = file_.name
        modules_def = self.toUnicode(unpacker.getText(), "utf8")
        files_def = self.toUnicode(unpacker.getText(), "utf8")
        sha_global = self.calculateShaGlobal()
        aqsql.AQSql.update("flupdates", ["actual"], [False], "1=1")
        aqsql.AQSql.insert(
            "flupdates",
            ["fecha", "hora", "nombre", "modulesdef", "filesdef", "shaglobal"],
            [
                now[: now.find("T")],
                now[(len(now) - (8)) :],
                file_name,
                modules_def,
                files_def,
                sha_global,
            ],
        )

    def warnLocalChanges(self, changes: Optional[Dict[str, Any]] = None) -> bool:
        """Show local changes warning."""

        if changes is None:
            changes = self.localChanges()
        if changes["size"] == 0:
            return True
        diag = qdialog.QDialog()
        diag.caption = self.translate("Detectados cambios locales")
        diag.setModal(True)
        txt = ""
        txt += self.translate("¡¡ CUIDADO !! DETECTADOS CAMBIOS LOCALES\n\n")
        txt += self.translate("Se han detectado cambios locales en los módulos desde\n")
        txt += self.translate("la última actualización/instalación de un paquete de módulos.\n")
        txt += self.translate("Si continua es posible que estos cambios sean sobreescritos por\n")
        txt += self.translate("los cambios que incluye el paquete que quiere cargar.\n\n")
        txt += "\n\n"
        txt += self.translate("Registro de cambios")
        lay = qvboxlayout.QVBoxLayout(diag)
        # lay.setMargin(6)
        # lay.setSpacing(6)
        lbl = qlabel.QLabel(diag)
        lbl.setText(txt)
        lbl.setAlignment(cast(QtCore.Qt.AlignmentFlag, QtCore.Qt.AlignmentFlag.AlignTop))
        lay.addWidget(lbl)
        ted = qtextedit.QTextEdit(diag)
        ted.setTextFormat(qtextedit.QTextEdit.LogText)
        ted.setAlignment(
            cast(
                QtCore.Qt.AlignmentFlag,
                QtCore.Qt.AlignmentFlag.AlignHCenter | QtCore.Qt.AlignmentFlag.AlignVCenter,
            )
        )
        ted.append(self.reportChanges(changes))
        lay.addWidget(ted)
        lbl2 = qlabel.QLabel(diag)
        lbl2.setText(self.translate("¿Que desea hacer?"))
        lbl2.setAlignment(cast(QtCore.Qt.AlignmentFlag, QtCore.Qt.AlignmentFlag.AlignTop))
        lay.addWidget(lbl2)
        lay2 = qhboxlayout.QHBoxLayout()
        # lay2.setMargin(6)
        # lay2.setSpacing(6)
        lay.addLayout(lay2)
        push_button_cancel = qpushbutton.QPushButton(diag)
        push_button_cancel.setText(self.translate("Cancelar"))
        push_button_accept = qpushbutton.QPushButton(diag)
        push_button_accept.setText(self.translate("continue"))
        lay2.addWidget(push_button_cancel)
        lay2.addWidget(push_button_accept)
        push_button_accept.clicked.connect(diag.accept)  # type: ignore [attr-defined]
        push_button_cancel.clicked.connect(diag.reject)  # type: ignore [attr-defined]
        if not application.PROJECT.app.platformName() != "offscreen":
            return False if (diag.exec() == 0) else True
        else:
            return True

    def xmlFilesDefBd(self) -> "QtXml.QDomDocument":
        """Return a QDomDocument with files definition."""

        doc = QtXml.QDomDocument("files_def")
        root = doc.createElement("files")
        doc.appendChild(root)
        qry = pnsqlquery.PNSqlQuery()
        qry.setSelect("idmodulo,nombre,contenido")
        qry.setFrom("flfiles")
        if not qry.exec_():
            return doc
        sha_sum = ""
        sha_sum_txt = ""
        sha_sum_bin = ""
        while qry.next():
            id_module = str(qry.value(0))
            if id_module == "sys":
                continue
            file_name = str(qry.value(1))
            ba_ = qbytearray.QByteArray()
            ba_.string = self.fromUnicode(str(qry.value(2)), "iso-8859-15")
            sha = ba_.sha1()
            node_file = doc.createElement("file")
            root.appendChild(node_file)
            node = doc.createElement("module")
            node_file.appendChild(node)
            node_text = doc.createTextNode(id_module)
            node.appendChild(node_text)
            node = doc.createElement("name")
            node_file.appendChild(node)
            node_text = doc.createTextNode(file_name)
            node.appendChild(node_text)
            if self.textPacking(file_name):
                node = doc.createElement("text")
                node_file.appendChild(node)
                node_text = doc.createTextNode(file_name)
                node.appendChild(node_text)
                node = doc.createElement("shatext")
                node_file.appendChild(node)
                node_text = doc.createTextNode(sha)
                node.appendChild(node_text)
                ba_ = qbytearray.QByteArray()
                ba_.string = sha_sum + sha
                sha_sum = ba_.sha1()
                ba_ = qbytearray.QByteArray()
                ba_.string = sha_sum_txt + sha
                sha_sum_txt = ba_.sha1()

        qry = pnsqlquery.PNSqlQuery()
        qry.setSelect("idmodulo,icono")
        qry.setFrom("flmodules")
        if qry.exec_():
            while qry.next():
                id_module = str(qry.value(0))
                if id_module == "sys":
                    continue
                file_name = utils_base.ustr(id_module, ".xpm")
                ba_ = qbytearray.QByteArray()
                ba_.string = str(qry.value(1))
                sha = ba_.sha1()
                node_file = doc.createElement("file")
                root.appendChild(node_file)
                node = doc.createElement("module")
                node_file.appendChild(node)
                node_text = doc.createTextNode(id_module)
                node.appendChild(node_text)
                node = doc.createElement("name")
                node_file.appendChild(node)
                node_text = doc.createTextNode(file_name)
                node.appendChild(node_text)
                if self.textPacking(file_name):
                    node = doc.createElement("text")
                    node_file.appendChild(node)
                    node_text = doc.createTextNode(file_name)
                    node.appendChild(node_text)
                    node = doc.createElement("shatext")
                    node_file.appendChild(node)
                    node_text = doc.createTextNode(sha)
                    node.appendChild(node_text)
                    ba_ = qbytearray.QByteArray()
                    ba_.string = sha_sum + sha
                    sha_sum = ba_.sha1()
                    ba_ = qbytearray.QByteArray()
                    ba_.string = sha_sum_txt + sha
                    sha_sum_txt = ba_.sha1()

        node = doc.createElement("shasum")
        node.appendChild(doc.createTextNode(sha_sum))
        root.appendChild(node)
        node = doc.createElement("shasumtxt")
        node.appendChild(doc.createTextNode(sha_sum_txt))
        root.appendChild(node)
        node = doc.createElement("shasumbin")
        node.appendChild(doc.createTextNode(sha_sum_bin))
        root.appendChild(node)
        return doc

    def loadModules(self, input_: Optional[Any] = None, warning_bakup: bool = True) -> bool:
        """Load modules from a package."""

        if input_ is None:
            util = flutil.FLUtil()
            dir_ = types.Dir(self.installPrefix())
            dir_.setCurrent()
            setting = (
                "scripts/sys/lastDirPackages_%s"
                % application.PROJECT.conn_manager.mainConn().DBName()
            )

            last_path = util.readSettingEntry(setting)
            path_tuple = QtWidgets.QFileDialog.getOpenFileName(
                QtWidgets.QApplication.focusWidget(),
                self.translate("scripts", "Seleccionar Eneboo/Abanq Package"),
                last_path,
                "Eneboo Package (*.eneboopkg);;Abanq Package (*.abanq)",
            )
            input_ = path_tuple[0]
            if input_:
                util.writeSettingEntry(setting, os.path.dirname(input_))

        return self.loadAbanQPackage(input_, warning_bakup)

    def loadAbanQPackage(self, input_: str, warning_bakup: bool = True) -> bool:
        """Load and process a Abanq/Eneboo package."""

        if input_:
            if warning_bakup and self.interactiveGUI():
                txt = ""
                txt += self.translate(
                    "Asegúrese de tener una copia de seguridad de todos los datos\n"
                )
                txt += self.translate(
                    "y de que  no hay ningun otro  usuario conectado a la base de\n"
                )
                txt += self.translate("datos mientras se realiza la carga.\n\n")
                txt += "\n\n"
                txt += self.translate("¿Desea continuar?")

                if messagebox.MessageBox.Yes != messagebox.MessageBox.warning(
                    txt, messagebox.MessageBox.No, messagebox.MessageBox.Yes
                ):
                    return False

            changes = self.localChanges()
            if changes["size"]:
                if not self.warnLocalChanges(changes):
                    return False

            unpacker = pnunpacker.PNUnpacker(input_)
            errors = unpacker.errorMessages()
            if len(errors):
                msg = self.translate("Hubo los siguientes errores al intentar cargar los módulos:")
                msg += "\n"

                for number in range(len(errors)):
                    msg += utils_base.ustr(errors[number], "\n")
                self.errorMsgBox(msg)
                return False

            unpacker.jump()
            unpacker.jump()
            unpacker.jump()

            if self.loadModulesDef(unpacker):
                if self.loadFilesDef(unpacker):
                    self.registerUpdate(input_)
                    if warning_bakup:
                        self.infoMsgBox(
                            self.translate("La carga de módulos se ha realizado con éxito.")
                        )
                    self.reinit()

                    tmp_var = flvar.FLVar()
                    tmp_var.set("mrproper", "dirty")
                    return True
                else:
                    self.errorMsgBox(
                        self.translate("No se ha podido realizar la carga de los módulos.")
                    )

        return False

    def loadFilesDef(self, document: "pnunpacker.PNUnpacker") -> bool:
        """Load files definition from a package to a QDomDocument."""

        files_definition = self.toUnicode(document.getText(), "utf8")
        doc = QtXml.QDomDocument()
        if not doc.setContent(files_definition):
            self.errorMsgBox(
                self.translate("Error XML al intentar cargar la definición de los ficheros.")
            )
            return False
        ok_ = True
        root = doc.firstChild()
        files = root.childNodes()
        flutil.FLUtil.createProgressDialog(self.translate("Registrando ficheros"), len(files))

        for number in range(len(files)):
            it_ = files.item(number)
            fil = {
                "id": it_.namedItem("name").toElement().text(),
                "skip": it_.namedItem("skip").toElement().text(),
                "module": it_.namedItem("module").toElement().text(),
                "text": it_.namedItem("text").toElement().text(),
                "shatext": it_.namedItem("shatext").toElement().text(),
                "binary": it_.namedItem("binary").toElement().text(),
                "shabinary": it_.namedItem("shabinary").toElement().text(),
            }
            flutil.FLUtil.setProgress(number)
            flutil.FLUtil.setLabelText(
                utils_base.ustr(self.translate("Registrando fichero"), " ", fil["id"])
            )
            if len(fil["id"]) == 0 or fil["skip"] == "true":
                continue
            if not self.registerFile(fil, document):
                self.errorMsgBox(
                    utils_base.ustr(self.translate("Error registrando el fichero"), " ", fil["id"])
                )
                ok_ = False
                break

        flutil.FLUtil.destroyProgressDialog()
        return ok_

    def registerFile(self, fil: Dict[str, Any], document: Any) -> bool:
        """Register a file in the database."""
        id_value: str = fil["id"]
        if id_value.endswith(".xpm"):
            cur = pnsqlcursor.PNSqlCursor("flmodules")
            if not cur.select(utils_base.ustr("idmodulo='", fil["module"], "'")) or not cur.first():
                return False

            cur.setModeAccess(aqsql.AQSql.Edit)
            cur.refreshBuffer()
            cur.setValueBuffer("icono", document.getText())
            return cur.commitBuffer()

        cur = pnsqlcursor.PNSqlCursor("flfiles")
        if not cur.select("nombre='%s'" % id_value):
            return False

        binary = id_value.endswith(".jasper")

        contenido = document.getBinary() if binary else document.getText()

        cur.setModeAccess((aqsql.AQSql.Edit if cur.first() else aqsql.AQSql.Insert))
        cur.refreshBuffer()
        cur.setValueBuffer("nombre", id_value)
        cur.setValueBuffer("idmodulo", fil["module"])
        cur.setValueBuffer("sha", fil["shabinary"] if binary else fil["shatext"])
        if binary:
            cur.setValueBuffer("contenido", "")
            cur.setValueBuffer("binario", contenido)
        else:
            if len(fil["text"]):
                encode = "iso-8859-15" if not id_value.endswith((".py")) else "UTF-8"
                try:
                    cur.setValueBuffer(
                        "contenido",
                        self.toUnicode(contenido, encode)
                        if not id_value.endswith(".py")
                        else contenido,
                    )
                except UnicodeEncodeError as error:
                    LOGGER.error(
                        "The %s file does not have the correct encode (%s)", id_value, encode
                    )
                    raise error

        return cur.commitBuffer()

    def checkProjectName(self, project_name: str) -> bool:
        """Return if te project name is valid."""
        if not project_name:
            project_name = ""
        db_project_name = flutil.FLUtil.readDBSettingEntry("projectname") or ""

        if project_name == db_project_name:
            return True

        if project_name and not db_project_name:
            return flutil.FLUtil.writeDBSettingEntry("projectname", project_name)

        txt = ""
        txt += self.translate("¡¡ CUIDADO !! POSIBLE INCOHERENCIA EN LOS MÓDULOS\n\n")
        txt += self.translate("Está intentando cargar un proyecto o rama de módulos cuyo\n")
        txt += self.translate("nombre difiere del instalado actualmente en la base de datos.\n")
        txt += self.translate("Es posible que la estructura de los módulos que quiere cargar\n")
        txt += self.translate(
            "sea completamente distinta a la instalada actualmente, y si continua\n"
        )
        txt += self.translate(
            "podría dañar el código, datos y la estructura de tablas de Eneboo.\n\n"
        )

        txt += self.translate("- Nombre del proyecto instalado: %s\n") % (str(db_project_name))
        txt += self.translate("- Nombre del proyecto a cargar: %s\n\n") % (str(project_name))
        txt += "\n\n"

        if not self.interactiveGUI():
            LOGGER.warning(txt)
            return False
        txt += self.translate("¿Desea continuar?")
        return messagebox.MessageBox.Yes == messagebox.MessageBox.warning(
            txt,
            messagebox.MessageBox.No,
            messagebox.MessageBox.Yes,
            messagebox.MessageBox.NoButton,
            "Pineboo",
        )

    def loadModulesDef(self, document: "pnunpacker.PNUnpacker") -> bool:
        """Return QDomDocument with modules definition."""

        modules_definition = self.toUnicode(document.getText(), "utf8")
        doc = QtXml.QDomDocument()
        if not doc.setContent(modules_definition):
            self.errorMsgBox(
                self.translate("Error XML al intentar cargar la definición de los módulos.")
            )
            return False
        root = doc.firstChild()
        if not self.checkProjectName(root.toElement().attribute("projectname", "")):
            return False
        ok_ = True
        modules = root.childNodes()
        flutil.FLUtil.createProgressDialog(self.translate("Registrando módulos"), len(modules))
        for number in range(len(modules)):
            it_ = modules.item(number)
            mod = {
                "id": it_.namedItem("name").toElement().text(),
                "alias": self.trTagText(it_.namedItem("alias").toElement().text()),
                "area": it_.namedItem("area").toElement().text(),
                "areaname": self.trTagText(it_.namedItem("areaname").toElement().text()),
                "version": it_.namedItem("version").toElement().text(),
            }
            flutil.FLUtil.setProgress(number)
            flutil.FLUtil.setLabelText(
                utils_base.ustr(self.translate("Registrando módulo"), " ", mod["id"])
            )
            if not self.registerArea(mod) or not self.registerModule(mod):
                self.errorMsgBox(
                    utils_base.ustr(self.translate("Error registrando el módulo"), " ", mod["id"])
                )
                ok_ = False
                break

        flutil.FLUtil.destroyProgressDialog()
        return ok_

    def registerArea(self, modules: Dict[str, Any]) -> bool:
        """Return True if the area is created or False."""
        cur = pnsqlcursor.PNSqlCursor("flareas")
        if not cur.select(utils_base.ustr("idarea = '", modules["area"], "'")):
            return False
        cur.setModeAccess((aqsql.AQSql.Edit if cur.first() else aqsql.AQSql.Insert))
        cur.refreshBuffer()
        cur.setValueBuffer("idarea", modules["area"])
        cur.setValueBuffer("descripcion", modules["areaname"])
        return cur.commitBuffer()

    def registerModule(self, modules: Dict[str, Any]) -> bool:
        """Return True if the module is created or False."""

        cur = pnsqlcursor.PNSqlCursor("flmodules")
        if not cur.select(utils_base.ustr("idmodulo='", modules["id"], "'")):
            return False
        cur.setModeAccess((aqsql.AQSql.Edit if cur.first() else aqsql.AQSql.Insert))
        cur.refreshBuffer()
        cur.setValueBuffer("idmodulo", modules["id"])
        cur.setValueBuffer("idarea", modules["area"])
        cur.setValueBuffer("descripcion", modules["alias"])
        cur.setValueBuffer("version", modules["version"])
        return cur.commitBuffer()

    def questionMsgBox(
        self,
        msg: str,
        key_remember: str = "",
        txt_remember: str = "",
        force_show: bool = True,
        txt_caption: str = "Pineboo",
        txt_yes: str = "Sí",
        txt_no: str = "No",
    ) -> Any:
        """Return a messagebox result."""

        key = "QuestionMsgBox/"
        value_remember = False
        if key_remember:
            value_remember = settings.SETTINGS.value(key + key_remember)
            if value_remember and not force_show:
                return messagebox.MessageBox.Yes
        if not self.interactiveGUI():
            return True
        diag = qdialog.QDialog()
        diag.caption = txt_caption
        diag.setModal(True)
        lay = qvboxlayout.QVBoxLayout(diag)
        # lay.setMargin(6)
        lay.setSpacing(6)
        lay2 = qhboxlayout.QHBoxLayout(diag)
        # lay2.setMargin(6)
        lay2.setSpacing(6)
        label_pix = qlabel.QLabel(diag)
        pixmap = aqs.AQS.pixmap_fromMimeSource("help_index.png")
        if pixmap:
            label_pix.setPixmap(pixmap)
            label_pix.setAlignment(aqs.AQS.AlignTop)
        lay2.addWidget(label_pix)
        lbl = qlabel.QLabel(diag)
        lbl.setText(msg)
        lbl.setAlignment(cast(QtCore.Qt.AlignmentFlag, aqs.AQS.AlignTop | aqs.AQS.WordBreak))
        lay2.addWidget(lbl)
        lay3 = qhboxlayout.QHBoxLayout(diag)
        # lay3.setMargin(6)
        lay3.setSpacing(6)
        push_button_yes = qpushbutton.QPushButton(diag)
        push_button_yes.setText(txt_yes if txt_yes else self.translate("Sí"))
        push_button_no = qpushbutton.QPushButton(diag)
        push_button_no.setText(txt_no if txt_no else self.translate("No"))
        lay3.addWidget(push_button_yes)
        lay3.addWidget(push_button_no)
        push_button_yes.clicked.connect(diag.accept)  # type: ignore [attr-defined]
        push_button_no.clicked.connect(diag.reject)  # type: ignore [attr-defined]
        check_remember = None
        if key_remember and txt_remember:
            check_remember = QtWidgets.QCheckBox(txt_remember, diag)
            check_remember.setChecked(value_remember)
            lay.addWidget(check_remember)

        if not application.PROJECT.app.platformName() == "offscreen":
            return messagebox.MessageBox.Yes

        ret = messagebox.MessageBox.No if (diag.exec() == 0) else messagebox.MessageBox.Yes
        if check_remember is not None:
            settings.SETTINGS.set_value(key + key_remember, check_remember.isChecked())
        return ret

    def exportModules(self) -> None:
        """Export modules."""

        dir_base_path = filedialog.FileDialog.getExistingDirectory(types.Dir.home)
        if not dir_base_path:
            return
        data_base_name = application.PROJECT.conn_manager.mainConn()._db_name
        dir_base_path = types.Dir.cleanDirPath(
            utils_base.ustr(
                dir_base_path,
                "/modulos_exportados_",
                data_base_name[data_base_name.rfind("/") + 1 :],
            )
        )
        dir_ = types.Dir()
        if not dir_.fileExists(dir_base_path):
            try:
                dir_.mkdir(dir_base_path)
            except Exception:
                error = traceback.format_exc()
                self.errorMsgBox(utils_base.ustr("", error))
                return

        else:
            self.warnMsgBox(
                dir_base_path + self.translate(" ya existe,\ndebe borrarlo antes de continuar")
            )
            return

        qry = pnsqlquery.PNSqlQuery()
        qry.setSelect("idmodulo")
        qry.setFrom("flmodules")
        if not qry.exec_() or qry.size() == 0:
            return
        pos = 0
        flutil.FLUtil.createProgressDialog(self.translate("Exportando módulos"), qry.size() - 1)
        while qry.next():
            id_module = qry.value(0)
            if id_module == "sys":
                continue
            flutil.FLUtil.setLabelText(id_module)
            pos += 1
            flutil.FLUtil.setProgress(pos)
            try:
                self.exportModule(id_module, dir_base_path)
            except Exception:
                error = traceback.format_exc()
                flutil.FLUtil.destroyProgressDialog()
                self.errorMsgBox(utils_base.ustr("", error))
                return

        db_project_name = flutil.FLUtil.readDBSettingEntry("projectname")
        if not db_project_name:
            db_project_name = ""
        if not db_project_name == "":
            doc = QtXml.QDomDocument()
            tag = doc.createElement("mvproject")
            tag.toElement().setAttribute("name", db_project_name)
            doc.appendChild(tag)
            try:
                types.FileStatic.write(
                    utils_base.ustr(dir_base_path, "/mvproject.xml"), doc.toString(2)
                )
            except Exception:
                error = traceback.format_exc()
                flutil.FLUtil.destroyProgressDialog()
                self.errorMsgBox(utils_base.ustr("", error))
                return

        flutil.FLUtil.destroyProgressDialog()
        self.infoMsgBox(self.translate("Módulos exportados en:\n") + dir_base_path)

    def xmlModule(self, id_module: str) -> "QtXml.QDomDocument":
        """Return xml data from a module."""
        qry = pnsqlquery.PNSqlQuery()
        qry.setSelect("descripcion,idarea,version")
        qry.setFrom("flmodules")
        qry.setWhere(utils_base.ustr("idmodulo='", id_module, "'"))
        doc = QtXml.QDomDocument("MODULE")
        if not qry.exec_() or not qry.next():
            return doc

        tag_module = doc.createElement("MODULE")
        doc.appendChild(tag_module)
        tag = doc.createElement("name")
        tag.appendChild(doc.createTextNode(id_module))
        tag_module.appendChild(tag)
        translate_noop = 'QT_TRANSLATE_NOOP("Eneboo","%s")'
        tag = doc.createElement("alias")
        tag.appendChild(doc.createTextNode(translate_noop % qry.value(0)))
        tag_module.appendChild(tag)
        id_area = qry.value(1)
        tag = doc.createElement("area")
        tag.appendChild(doc.createTextNode(id_area))
        tag_module.appendChild(tag)
        area_name = utils_db.sql_select(
            "flareas", "descripcion", utils_base.ustr("idarea='", id_area, "'")
        )
        tag = doc.createElement("areaname")
        tag.appendChild(doc.createTextNode(translate_noop % area_name))
        tag_module.appendChild(tag)
        tag = doc.createElement("entryclass")
        tag.appendChild(doc.createTextNode(id_module))
        tag_module.appendChild(tag)
        tag = doc.createElement("version")
        tag.appendChild(doc.createTextNode(qry.value(2)))
        tag_module.appendChild(tag)
        tag = doc.createElement("icon")
        tag.appendChild(doc.createTextNode(utils_base.ustr(id_module, ".xpm")))
        tag_module.appendChild(tag)
        return doc

    def fileWriteIso(self, file_name: str, content: str) -> None:
        """Write data into a file with ISO-8859-15 encode."""

        file_iso = types.File(file_name, "ISO8859-15")
        file_iso.write(content.encode("ISO8859-15", "ignore"))
        file_iso.close()

    def fileWriteUtf8(self, file_name: str, content: str) -> None:
        """Write data into a file with UTF-8 encode."""

        file_utf = types.File(file_name, "UTF-8")
        file_utf.write(content)
        file_utf.close()

    def exportModule(self, id_module: str, dir_base_path: str) -> None:
        """Export a module to a directory."""

        dir_ = types.Dir()
        dir_path = types.Dir.cleanDirPath(utils_base.ustr(dir_base_path, "/", id_module))
        if not dir_.fileExists(dir_path):
            dir_.mkdir(dir_path)
        for name in ["/forms", "/scripts", "/queries", "/tables", "/reports", "/translations"]:
            if not dir_.fileExists("%s%s" % (dir_path, name)):
                dir_.mkdir("%s%s" % (dir_path, name))
        xml_module = self.xmlModule(id_module)
        self.fileWriteIso(utils_base.ustr(dir_path, "/", id_module, ".mod"), xml_module.toString(2))
        xpm_module = utils_db.sql_select(
            "flmodules", "icono", utils_base.ustr("idmodulo='", id_module, "'")
        )
        self.fileWriteIso(utils_base.ustr(dir_path, "/", id_module, ".xpm"), xpm_module)
        qry = pnsqlquery.PNSqlQuery()
        qry.setSelect("nombre,contenido")
        qry.setFrom("flfiles")
        qry.setWhere(utils_base.ustr("idmodulo='", id_module, "'"))
        if qry.exec_():
            while qry.next():
                name = qry.value(0)
                content = qry.value(1)
                type_ = name[name.rfind(".") :]
                if type_ == ".xml":
                    self.fileWriteIso(utils_base.ustr(dir_path, "/", name), content)
                elif type_ == ".ui":
                    self.fileWriteIso(utils_base.ustr(dir_path, "/forms/", name), content)
                elif type_ == ".qs":
                    self.fileWriteIso(utils_base.ustr(dir_path, "/scripts/", name), content)
                elif type_ == ".qry":
                    self.fileWriteIso(utils_base.ustr(dir_path, "/queries/", name), content)
                elif type_ == ".mtd":
                    self.fileWriteIso(utils_base.ustr(dir_path, "/tables/", name), content)
                elif type_ in (".kut", ".ar", ".jrxml", ".svg"):
                    self.fileWriteIso(utils_base.ustr(dir_path, "/reports/", name), content)
                elif type_ == ".ts":
                    self.fileWriteIso(utils_base.ustr(dir_path, "/translations/", name), content)
                elif type_ == ".py":
                    self.fileWriteUtf8(utils_base.ustr(dir_path, "/scripts/", name), content)

    def importModules(self, warning_bakup: bool = True) -> None:
        """Import modules from a directory."""

        if warning_bakup and self.interactiveGUI():
            txt = ""
            txt += self.translate("Asegúrese de tener una copia de seguridad de todos los datos\n")
            txt += self.translate("y de que  no hay ningun otro  usuario conectado a la base de\n")
            txt += self.translate("datos mientras se realiza la importación.\n\n")
            txt += self.translate("Obtenga soporte en")
            txt += " http://www.infosial.com\n(c) InfoSiAL S.L."
            txt += "\n\n"
            txt += self.translate("¿Desea continuar?")
            if messagebox.MessageBox.Yes != messagebox.MessageBox.warning(
                txt, messagebox.MessageBox.No, messagebox.MessageBox.Yes
            ):
                return

        key = utils_base.ustr("scripts/sys/modLastDirModules_", self.nameBD())
        dir_ant = settings.SETTINGS.value(key)

        dir_modules = filedialog.FileDialog.getExistingDirectory(
            str(dir_ant) if dir_ant else ".", self.translate("Directorio de Módulos")
        )
        if not dir_modules:
            return
        dir_modules = types.Dir.cleanDirPath(dir_modules)
        dir_modules = types.Dir.convertSeparators(dir_modules)
        QtCore.QDir.setCurrent(dir_modules)  # change current directory
        modified_files = self.selectModsDialog(flutil.FLUtil.findFiles(dir_modules, "*.mod", False))
        flutil.FLUtil.createProgressDialog(self.translate("Importando"), len(modified_files))
        flutil.FLUtil.setProgress(1)

        for number, value in enumerate(modified_files):
            flutil.FLUtil.setLabelText(value)
            flutil.FLUtil.setProgress(number)
            if not self.importModule(value):
                self.errorMsgBox(self.translate("Error al cargar el módulo:\n") + value)
                break

        flutil.FLUtil.destroyProgressDialog()
        flutil.FLUtil.writeSettingEntry(key, dir_modules)
        self.infoMsgBox(self.translate("Importación de módulos finalizada."))
        AQTimer.singleShot(0, self.reinit)  # type: ignore [arg-type] # noqa: F821

    def selectModsDialog(self, modified_files: List = []) -> "types.Array":
        """Select modules dialog."""

        dialog_ = dialog.Dialog()
        dialog_.okButtonText = self.translate("Aceptar")
        dialog_.cancelButtonText = self.translate("Cancelar")
        bgroup = QtWidgets.QGroupBox()
        bgroup.setTitle(self.translate("Seleccione módulos a importar"))
        dialog_.add(bgroup)
        res = types.Array()
        check_box = types.Array()

        for number, item in enumerate(modified_files):
            check_box[number] = QtWidgets.QCheckBox()

            check_box[number].text = item
            check_box[number].checked = True

        idx = 0
        if self.interactiveGUI() and dialog_.exec_():
            for number, item in enumerate(modified_files):
                if check_box[number].checked:
                    res[idx] = item
                    idx += 1

        return res

    def importModule(self, module_path: str) -> bool:
        """Import a module specified by name."""
        try:
            with open(module_path, "r", encoding="ISO8859-15") as file_module:
                content_module = file_module.read()
        except Exception:
            error = traceback.format_exc()
            self.errorMsgBox(utils_base.ustr(self.translate("Error leyendo fichero."), "\n", error))
            return False
        mod_folder = os.path.dirname(module_path)
        mod = None
        xml_module = QtXml.QDomDocument()
        if xml_module.setContent(content_module):
            node_module = xml_module.namedItem("MODULE")
            if not node_module:
                self.errorMsgBox(self.translate("Error en la carga del fichero xml .mod"))
                return False
            mod = {
                "id": (node_module.namedItem("name").toElement().text()),
                "alias": (self.trTagText(node_module.namedItem("alias").toElement().text())),
                "area": (node_module.namedItem("area").toElement().text()),
                "areaname": (self.trTagText(node_module.namedItem("areaname").toElement().text())),
                "version": (node_module.namedItem("version").toElement().text()),
            }
            if not self.registerArea(mod) or not self.registerModule(mod):
                self.errorMsgBox(
                    utils_base.ustr(self.translate("Error registrando el módulo"), " ", mod["id"])
                )
                return False

            for ext in [
                "*.xml",
                "*.ui",
                "*.qs",
                "*.py",
                "*.qry",
                "*.mtd",
                "*.kut",
                "*.ar",
                "*.jrxml",
                "*.svg",
                "*.ts",
            ]:
                if not self.importFiles(mod_folder, ext, mod["id"]):
                    return False

        else:
            self.errorMsgBox(self.translate("Error en la carga del fichero xml .mod"))
            return False

        return True

    def importFiles(self, dir_path_: str, ext: str, id_module_: str) -> bool:
        """Import files with a exension from a path."""
        ok_ = True
        util = flutil.FLUtil()
        list_files_ = util.findFiles(dir_path_, ext, False)
        util.createProgressDialog(self.translate("Importando"), len(list_files_))
        util.setProgress(1)

        for number, value in enumerate(list_files_):
            util.setLabelText(value)
            util.setProgress(number)
            if not self.importFile(value, id_module_):
                self.errorMsgBox(self.translate("Error al cargar :\n") + value)
                ok_ = False
                break

        util.destroyProgressDialog()
        return ok_

    def importFile(self, file_path_: str, id_module_: str) -> bool:
        """Import a file from a path."""
        file_ = types.File(file_path_)
        content = ""
        try:
            file_.open(types.File.ReadOnly)
            content = str(file_.read())
        except Exception:
            error = traceback.format_exc()
            self.errorMsgBox(utils_base.ustr(self.translate("Error leyendo fichero."), "\n", error))
            return False

        ok_ = True
        name: str = file_.name
        if (
            not flutil.FLUtil.isFLDefFile(content)
            and not name.endswith((".qs", ".py", ".ar", ".svg"))
        ) or name.endswith("untranslated.ts"):
            return ok_
        cur = pnsqlcursor.PNSqlCursor("flfiles")
        cur.select(utils_base.ustr("nombre = '%s'" % name))
        ba_ = qbytearray.QByteArray()
        ba_.string = content
        sha_count = ba_.sha1()
        copy_content = ""

        if cur.first():
            copy_content = cur.valueBuffer("contenido")
            cur.setModeAccess(aqsql.AQSql.Edit)
        else:
            cur.setModeAccess(aqsql.AQSql.Insert)

        if name.endswith(".ar"):
            if not self.importReportAr(file_path_, id_module_, content):
                return True

        cur.refreshBuffer()
        cur.setValueBuffer("nombre", name)
        cur.setValueBuffer("idmodulo", id_module_)
        cur.setValueBuffer("sha", sha_count)
        cur.setValueBuffer("contenido", content)
        ok_ = cur.commitBuffer()

        if ok_ and copy_content:
            date_ = str(types.Date())
            cur.setModeAccess(aqsql.AQSql.Insert)
            cur.refreshBuffer()
            cur.setValueBuffer("nombre", "%s%s" % (name, date_))
            cur.setValueBuffer("idmodulo", id_module_)
            cur.setValueBuffer("contenido", copy_content)
            ok_ = cur.commitBuffer()

        return ok_

    def importReportAr(self, file_path_: str, id_module_: str, content: str) -> bool:
        """Import a report file, convert and install."""

        from pineboolib.application.safeqsa import SafeQSA

        if (
            not self.isLoadedModule("flar2kut")
            or settings.SETTINGS.value("scripts/sys/conversionAr") != "true"
        ):
            return False

        content = self.toUnicode(content, "UTF-8")
        content = SafeQSA.root_module("flar2kut").iface.pub_ar2kut(content)
        file_path_ = utils_base.ustr(file_path_[0 : len(file_path_) - 3], ".kut")
        if content:
            local_encoding = settings.SETTINGS.value("scripts/sys/conversionArENC")
            if not local_encoding:
                local_encoding = "ISO-8859-15"
            content = self.fromUnicode(content, local_encoding)
            file_ = types.FileStatic()
            try:
                file_.write(file_path_, content)
            except Exception:
                error = traceback.format_exc()
                self.errorMsgBox(
                    utils_base.ustr(self.translate("Error escribiendo fichero."), "\n", error)
                )
                return False

            return self.importFile(file_path_, id_module_)

        return False

    def runTransaction(self, function: Callable, optional_params: Dict[str, Any]) -> Any:
        """Run a Transaction."""
        roll_back_: bool = False
        error_msg_: str = ""
        valor_: Any

        db_ = application.PROJECT.conn_manager.useConn("default")

        # Create Transaction.
        db_.transaction()
        db_._transaction_level += 1

        if self.interactiveGUI():
            QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.CursorShape.WaitCursor)

        try:
            valor_ = function(optional_params)
            if "errorMsg" in optional_params:
                error_msg_ = optional_params["errorMsg"]

            if not valor_:
                roll_back_ = True

        except Exception:
            error = traceback.format_exc(limit=-6, chain=False)
            roll_back_ = True
            valor_ = False
            if error_msg_ == "":
                error_msg_ = self.translate("Error al ejecutar la función")
            raise Exception("%s: %s" % (error_msg_, error))

        db_._transaction_level -= 1

        if roll_back_:  # do RollBack
            if error_msg_ != "":
                self.warnMsgBox(error_msg_)

            db_.rollback()

        else:  # do Commit
            db_.commit()

        if self.interactiveGUI():
            aqs.AQS.Application_restoreOverrideCursor()

        return valor_

    @decorators.deprecated
    def qsaExceptions(self) -> None:
        """Return QSA exceptions found."""

        return
        # return application.PROJECT.conn_manager.qsaExceptions()

    def serverTime(self) -> str:
        """Return time from database."""
        conn = application.PROJECT.conn_manager.useConn("default")
        qry_result = conn.execute_query("SELECT current_time")
        result = ""
        if qry_result is not None:
            line = qry_result.fetchone()
            if line is not None:
                result = str(line[0])

        return result

    def localChanges(self) -> Dict[str, Any]:
        """Return xml with local changes."""
        ret = {}
        ret["size"] = 0
        str_xml_update = utils_db.quick_sql_select(
            "flupdates",
            "filesdef",
            "actual=%s"
            % application.PROJECT.conn_manager.default().formatValue("bool", True, False),
        )
        if not str_xml_update:
            return ret
        document_update = QtXml.QDomDocument()
        if not document_update.setContent(str_xml_update):
            self.errorMsgBox(
                self.translate("Error XML al intentar cargar la definición de los ficheros.")
            )
            return ret
        document_db = self.xmlFilesDefBd()

        return self.diffXmlFilesDef(document_db, document_update)

    @classmethod
    def interactiveGUI(cls) -> str:
        """Return interactiveGUI."""

        return application.PROJECT.conn_manager.mainConn().interactiveGUI()

    def getWidgetList(self, container: str, control_name: str) -> str:
        """Get widget list from a widget."""

        obj_class: Any = None
        if control_name == "FLFieldDB":
            obj_class = flfielddb.FLFieldDB
        elif control_name == "FLTableDB":
            obj_class = fltabledb.FLTableDB
        elif control_name == "Button":
            control_name = "QPushButton"

        if obj_class is None:
            obj_class = getattr(QtWidgets, control_name, None)

        if obj_class is None:
            raise Exception("obj_class is empty!")

        widget: Optional[Union["flformrecorddb.FLFormRecordDB", "flformdb.FLFormDB"]] = None
        action = None
        conn = application.PROJECT._conn_manager
        if conn is None:
            raise Exception("conn is empty!")

        if container[0:10] == "formRecord":
            action_ = container[10:]
            action = conn.manager().action(action_)
            if action.formRecord():
                widget = conn.managerModules().createFormRecord(action)
        elif container[0:10] == "formSearch":
            action_ = container[10:]
            action = conn.manager().action(action_)
            if action.form():
                widget = conn.managerModules().createForm(action)
        else:
            action_ = container[4:]
            action = conn.manager().action(action_)
            if action.form():
                widget = conn.managerModules().createForm(action)

        if widget is None:
            return ""

        object_list = widget.findChildren(obj_class)
        retorno_: str = ""
        for obj in object_list:
            name_ = obj.objectName()
            if name_ == "":
                continue

            if control_name == "FLFieldDB":
                field_table_ = cast(flfielddb.FLFieldDB, obj).tableName()
                if field_table_ and field_table_ != action.table():
                    continue
                retorno_ += "%s/%s*" % (name_, cast(flfielddb.FLFieldDB, obj).fieldName())
            elif control_name == "FLTableDB":
                retorno_ += "%s/%s*" % (name_, cast(fltabledb.FLTableDB, obj).tableName())
            elif control_name in ["QPushButton", "Button"]:
                if name_ in ["pushButtonDB", "pbAux", "qt_left_btn", "qt_right_btn"]:
                    continue
                retorno_ += "%s/%s*" % (name_, obj.objectName())
            else:
                if name_ in [
                    "textLabelDB",
                    "componentDB",
                    "tab_pages",
                    "editor",
                    "FrameFind",
                    "TextLabelSearch",
                    "TextLabelIn",
                    "lineEditSearch",
                    "in-combo",
                    "voidTable",
                ]:
                    continue
                if isinstance(obj, QtWidgets.QGroupBox):
                    retorno_ += "%s/%s*" % (name_, obj.title())
                else:
                    retorno_ += "%s/*" % (name_)

        return retorno_


class AbanQDbDumper(QtCore.QObject):
    """AbanqDbDumper class."""

    SEP_CSV = "\u00b6"
    db_: "iconnection.IConnection"
    _show_gui: bool
    _dir_base: str
    _file_name: str
    widget_: "QtWidgets.QDialog"
    _label_dir_base: "qlabel.QLabel"
    pushbutton_change_dir: "qpushbutton.QPushButton"
    _ted_log: "qtextedit.QTextEdit"
    pb_init_dump: "qpushbutton.QPushButton"
    state_: "types.Array"
    _fun_log: "Callable"
    proc_: "process.Process"

    def __init__(
        self,
        db_: Optional["iconnection.IConnection"] = None,
        dir_base: Optional[str] = None,
        show_gui: bool = True,
        fun_log: Optional[Callable] = None,
    ):
        """Inicialize."""

        super().__init__()

        self._fun_log = self.addLog if fun_log is None else fun_log  # type: ignore

        self.db_ = application.PROJECT.aq_app.db().mainConn() if db_ is None else db_
        self._show_gui = show_gui
        self._dir_base = types.Dir.home if dir_base is None else dir_base

        self._file_name = self.genFileName()
        self.encoding = sys.getdefaultencoding()
        self.state_ = types.Array()

    def init(self) -> None:
        """Inicialize dump dialog."""
        if self._show_gui:
            self.buildGui()
            self.widget_.exec()

    def buildGui(self) -> None:
        """Build a Dialog for database dump."""
        self.widget_ = QtWidgets.QDialog(application.PROJECT.main_window)
        self.widget_.setWindowTitle(SysType.translate("Copias de seguridad"))
        self.widget_.resize(800, 600)

        lay = qvboxlayout.QVBoxLayout(self.widget_)
        lay.setContentsMargins(6, 6, 6, 6)
        lay.setSpacing(6)

        frm = QtWidgets.QFrame(self.widget_)
        frm.setFrameShape(QtWidgets.QFrame.Shape.Box)
        frm.setLineWidth(1)
        frm.setFrameShadow(QtWidgets.QFrame.Shadow.Plain)

        # lay_frame = qvboxlayout.QVBoxLayout(frm, 6, 6)
        lay_frame = qvboxlayout.QVBoxLayout(frm)
        lay_frame.setContentsMargins(6, 6, 6, 6)
        lay_frame.setSpacing(6)
        lbl = qlabel.QLabel(frm)
        lbl.setText(
            SysType.translate("Driver: %s")
            % (str(self.db_.driverNameToDriverAlias(self.db_.driverName())))
        )
        lbl.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        lay_frame.addWidget(lbl)
        lbl = qlabel.QLabel(frm)
        lbl.setText(SysType.translate("Base de datos: %s") % (str(self.db_.database())))
        lbl.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        lay_frame.addWidget(lbl)
        lbl = qlabel.QLabel(frm)
        lbl.setText(SysType.translate("Host: %s") % (str(self.db_.host())))
        lbl.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        lay_frame.addWidget(lbl)
        lbl = qlabel.QLabel(frm)
        lbl.setText(SysType.translate("Puerto: %s") % (str(self.db_.port())))
        lbl.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        lay_frame.addWidget(lbl)
        lbl = qlabel.QLabel(frm)
        lbl.setText(SysType.translate("Usuario: %s") % (str(self.db_.user())))
        lbl.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        lay_frame.addWidget(lbl)
        lay_aux = qhboxlayout.QHBoxLayout()
        lay_frame.addLayout(lay_aux)
        self._label_dir_base = qlabel.QLabel(frm)
        self._label_dir_base.setText(
            SysType.translate("Directorio Destino: %s") % (str(self._dir_base))
        )
        self._label_dir_base.setAlignment(QtCore.Qt.AlignmentFlag.AlignVCenter)
        lay_aux.addWidget(self._label_dir_base)
        self.pushbutton_change_dir = qpushbutton.QPushButton(SysType.translate("Cambiar"))
        self.pushbutton_change_dir.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Preferred
        )

        self.pushbutton_change_dir.clicked.connect(  # type: ignore [attr-defined]
            self.changeDirBase
        )
        lay_aux.addWidget(self.pushbutton_change_dir)
        lay.addWidget(frm)
        self.pb_init_dump = qpushbutton.QPushButton(SysType.translate("INICIAR COPIA"))
        self.pb_init_dump.clicked.connect(self.initDump)  # type: ignore [attr-defined]
        lay.addWidget(self.pb_init_dump)
        lbl = qlabel.QLabel(self.widget_)
        lbl.setText("Log:")
        lay.addWidget(lbl)
        self._ted_log = qtextedit.QTextEdit(self.widget_)
        self._ted_log.setTextFormat(qtextedit.QTextEdit.LogText)
        self._ted_log.setAlignment(
            cast(
                QtCore.Qt.AlignmentFlag,
                QtCore.Qt.AlignmentFlag.AlignLeft | QtCore.Qt.AlignmentFlag.AlignVCenter,
            )
        )
        lay.addWidget(self._ted_log)

    def initDump(self) -> None:
        """Inicialize dump."""

        gui = self._show_gui and self.widget_ is not None
        if gui:
            self.widget_.setEnabled(False)
        self.dumpDatabase()
        if gui:
            self.widget_.setEnabled(True)
            SysType.infoMsgBox(self.state_.msg)
            if self.state_.ok:
                self.widget_.close()

    def genFileName(self) -> str:
        """Return a file name."""
        now = types.Date()
        time_stamp = str(now)
        reg_exp = ["-", ":"]
        # reg_exp.global_ = True
        for item in reg_exp:
            time_stamp = time_stamp.replace(item, "")

        file_name = "%s/dump_%s_%s" % (self._dir_base, self.db_.database(), time_stamp)
        file_name = types.Dir.cleanDirPath(file_name)
        file_name = types.Dir.convertSeparators(file_name)
        return file_name

    def changeDirBase(self, dir_: Optional[str] = None) -> None:
        """Change base dir."""

        dir_base_path = dir_
        if not dir_base_path:
            dir_base_path = filedialog.FileDialog.getExistingDirectory(self._dir_base)
            if not dir_base_path:
                return
        self._dir_base = dir_base_path
        if self._show_gui and self._label_dir_base is not None:
            self._label_dir_base.setText(
                SysType.translate("Directorio Destino: %s") % (str(self._dir_base))
            )
        self._file_name = self.genFileName()

    def addLog(self, msg: str) -> None:
        """Add a text to log."""

        if self._show_gui and self._ted_log is not None:
            self._ted_log.append(msg)
        else:
            LOGGER.warning(msg)

    def setState(self, ok_: int, msg: str) -> None:
        """Set state."""

        self.state_.ok = ok_
        self.state_.msg = msg

    def state(self) -> types.Array:
        """Return state."""

        return self.state_

    def launchProc(self, command: List[str]) -> str:
        """Return the result from a Launched command."""
        self.proc_ = process.Process()
        self.proc_.setProgram(command[0])
        self.proc_.setArguments(command[1:])

        self.proc_.readyReadStandardOutput.connect(self.readFromStdout)  # type: ignore [attr-defined]
        self.proc_.readyReadStandardError.connect(self.readFromStdout)  # type: ignore [attr-defined]

        self.proc_.start(QtCore.QIODeviceBase.OpenModeFlag.ReadOnly)

        while self.proc_.running:
            SysType.processEvents()

        return self.proc_.exitcode() == self.proc_.ExitStatus.NormalExit.value

    def readFromStdout(self) -> None:
        """Read data from stdOutput."""
        while self.proc_.canReadLine():
            text = self.proc_.readLine().decode(self.encoding)
            if text not in (None, ""):
                self._fun_log(text)

    def dumpDatabase(self) -> bool:
        """Dump database to target specified by sql driver class."""

        driver = self.db_.driverName()
        type_db = 0
        if driver.find("PSQL") > -1:
            type_db = 1
        elif driver.find("MYSQL") > -1:
            type_db = 2

        if type_db == 0:
            self.setState(
                False,
                SysType.translate("Este tipo de base de datos no soporta el volcado a disco."),
            )
            self._fun_log(self.state_.msg)
            self.dumpAllTablesToCsv()
            return False
        file = types.File(self._file_name)  # noqa
        try:
            if not os.path.exists(self._file_name):
                dir_ = types.Dir(self._file_name)  # noqa

        except Exception:
            error = traceback.format_exc()
            self.setState(False, utils_base.ustr("", error))
            self._fun_log(self.state_.msg)
            return False

        ok_ = True
        if type_db == 1:
            ok_ = self.dumpPostgreSQL()

        if type_db == 2:
            ok_ = self.dumpMySQL()

        if not ok_:
            self.dumpAllTablesToCsv()
        if not ok_:
            self.setState(
                False, SysType.translate("No se ha podido realizar la copia de seguridad.")
            )
            self._fun_log(self.state_.msg)
        else:
            self.setState(
                True,
                SysType.translate("Copia de seguridad realizada con éxito en:\n%s.sql")
                % (str(self._file_name)),
            )
            self._fun_log(self.state_.msg)

        return ok_

    def dumpPostgreSQL(self) -> bool:
        """Dump database to PostgreSql file."""

        from pineboolib.core import system as system_mod

        pg_dump: str = "pg_dump"
        command: List[str] = []
        file_name = "%s.sql" % self._file_name

        system_mod.System.setenv("PGPASSWORD", self.db_.returnword())

        if SysType.osName() == "WIN32":
            pg_dump += ".exe"

        command = [
            pg_dump,
            "-f",
            file_name,
            "-h",
            self.db_.host(),
            "-p",
            str(self.db_.port()),
            "-U",
            self.db_.user(),
            self.db_.database().DBName(),
        ]

        if not self.launchProc(command):
            self.setState(
                False,
                SysType.translate("No se ha podido volcar la base de datos a disco.\n")
                + SysType.translate("Es posible que no tenga instalada la herramienta ")
                + pg_dump,
            )
            self._fun_log(self.state_.msg)
            return False
        self.setState(True, "")
        return True

    def dumpMySQL(self) -> bool:
        """Dump database to MySql file."""

        my_dump: str = "mysqldump"
        command: List[str]
        file_name = utils_base.ustr(self._file_name, ".sql")

        if SysType.osName() == "WIN32":
            my_dump += ".exe"
        command = [
            my_dump,
            "-v",
            utils_base.ustr("--result-file=", file_name),
            utils_base.ustr("--host=", self.db_.host()),
            utils_base.ustr("--port=", self.db_.port()),
            utils_base.ustr("--password=", self.db_.returnword()),
            utils_base.ustr("--user=", self.db_.user()),
            str(self.db_.database()),
        ]

        if not self.launchProc(command):
            self.setState(
                False,
                SysType.translate("No se ha podido volcar la base de datos a disco.\n")
                + SysType.translate("Es posible que no tenga instalada la herramienta ")
                + my_dump,
            )
            self._fun_log(self.state_.msg)
            return False
        self.setState(True, "")
        return True

    def dumpTableToCsv(self, table: str, dir_base: str) -> bool:
        """Dump a table to a CSV."""

        file_name = utils_base.ustr(dir_base, table, ".csv")
        file_ = types.File(file_name)
        if not file_.open(types.File.WriteOnly):
            return False
        ts_ = QtCore.QTextStream(file_.ioDevice())
        # ts_.setCodec(aqs.AQS.TextCodec_codecForName(u"utf8"))
        qry = pnsqlquery.PNSqlQuery()
        qry.setSelect(utils_base.ustr(table, ".*"))
        qry.setFrom(table)
        if not qry.exec_():
            return False

        rec = str("%s" % self.SEP_CSV).join(qry.fieldList())

        ts_.device().write(utils_base.ustr(rec, "\n").encode())  # type: ignore [union-attr]
        # ts.opIn(utils_base.ustr(rec, u"\n"))
        flutil.FLUtil.createProgressDialog(
            SysType.translate("Haciendo copia en CSV de ") + table, qry.size()
        )
        pos = 0
        while qry.next():
            values = []
            for field_name in qry.fieldList():
                values.append(str(qry.value(field_name)))

            rec = str("%s" % self.SEP_CSV).join(values)

            ts_.device().write(utils_base.ustr(rec, "\n").encode())  # type: ignore [union-attr]
            pos += 1
            flutil.FLUtil.setProgress(pos)

        file_.close()
        flutil.FLUtil.destroyProgressDialog()
        return True

    def dumpAllTablesToCsv(self) -> bool:
        """Dump all tables to a csv files."""
        tables = self.db_.tables(aqsql.AQSql.TableType.Tables)
        dir_ = types.Dir(self._file_name)
        dir_.mkdir()
        dir_base = types.Dir.convertSeparators(utils_base.ustr(self._file_name, "/"))
        # i = 0
        # while_pass = True
        for table_ in tables:
            self.dumpTableToCsv(table_, dir_base)
        return True



"""Main module for starting up Pineboo."""

from pineboolib import application, logging
from pineboolib.application.utils import external

from pineboolib.core import settings

from pineboolib.loader import dgi as dgi_module
from pineboolib.loader import connection
from pineboolib.core.utils import utils_base


import gc
import sys
import shutil
import os


from typing import List, Type, Optional, TYPE_CHECKING


if TYPE_CHECKING:
    from pineboolib.loader import projectconfig  # noqa: F401 # pragma: no cover
    import optparse  # noqa: F401 # pragma: no cover
    from PyQt6 import QtWidgets  # type: ignore[import] # pragma: no cover
    from types import TracebackType  # pragma: no cover

LOGGER = logging.get_logger(__name__)


def startup_no_x() -> None:
    """Start Pineboo with no GUI."""
    startup(enable_gui=False)


def startup_framework(conn: Optional["projectconfig.ProjectConfig"] = None) -> None:
    """Start Pineboo project like framework."""
    if conn is None:
        raise Exception("conn is empty!")

    import pyfiglet  # type: ignore

    qapp = call_qapplication(sys.argv + ["-platform", "offscreen"])
    init_logging(True, application.FRAMEWORK_DEBUG_LEVEL)
    init_cli(catch_ctrl_c=False)

    LOGGER.info(
        pyfiglet.figlet_format(
            "\nPINEBOO %s " % application.PROJECT.load_version(), font="starwars"
        )
    )

    application.PROJECT.set_app(qapp)
    dgi = dgi_module.load_dgi("qt", None)
    application.PROJECT.init_dgi(dgi)
    application.PROJECT.aq_app._inicializing = False

    LOGGER.info("STARTUP_FRAMEWORK:(1/7) Setting profile data.")
    conn_ = connection.connect_to_db(conn)
    LOGGER.info("STARTUP_FRAMEWORK:(2/7) Establishing connection.")

    if not application.PROJECT.init_conn(connection=conn_):
        raise Exception("No main connection was established. Aborting Pineboo load.")

    external.load_project_config_file()

    _initialize_data(True)


def startup(enable_gui: Optional[bool] = None) -> None:
    """Start up pineboo."""
    # FIXME: No hemos cargado pineboo aún. No se pueden usar métodos internos.
    from pineboolib.core.utils import check_dependencies
    from pineboolib.loader import options as options_module

    if not check_dependencies.check_dependencies_cli(
        {"ply": "python3-ply", "PyQt6.QtCore": "python3-PyQt6", "Python": "Python"}
    ):
        sys.exit(32)

    min_python = (3, 6)
    if sys.version_info < min_python:
        sys.exit("Python %s.%s or later is required.\n" % min_python)

    options = options_module.parse_options()

    if options.pineboo_version:
        print("Pineboo %s." % application.PINEBOO_VER)
        sys.exit(0)

    print(chr(27) + "[2J")

    if enable_gui is not None:
        options.enable_gui = enable_gui
    trace_loggers: List[str] = []
    if options.trace_loggers:
        trace_loggers = options.trace_loggers.split(",")

    init_logging(logtime=options.log_time, loglevel=options.loglevel, trace_loggers=trace_loggers)

    if options.project_name:
        application.PROJECT_NAME = options.project_name

    if options.external:
        if not os.path.exists(options.external):
            LOGGER.error("External: %s is not a valid directory", options.external)
            sys.exit(1)
        LOGGER.info("External: Adding %s to sys.path", options.external)
        sys.path.insert(0, options.external)
        application.EXTERNAL_FOLDER = options.external
        external.load_project_config_file()

    ret = exec_main_with_profiler(options) if options.enable_profiler else exec_main(options)

    gc.collect()
    LOGGER.info("Closing Pineboo...")
    sys.exit(ret if ret else 0)


def init_logging(
    logtime: bool = False, loglevel: int = logging.INFO, trace_loggers: List[str] = []
) -> None:
    """Initialize pineboo logging."""

    # ---- LOGGING -----
    log_format = "%(levelname)-8s: %(thread)s: %(name)s:%(lineno)d: %(message)s"

    if logtime:
        log_format = "%(asctime)s - %(levelname)-8s: %(thread)s: %(name)s:%(lineno)d: %(message)s"

    app_loglevel = logging.TRACE if trace_loggers else loglevel

    import coloredlogs  # type: ignore [import]

    coloredlogs.DEFAULT_LOG_LEVEL = app_loglevel
    coloredlogs.DEFAULT_LOG_FORMAT = log_format
    # 'black', 'blue', 'cyan', 'green', 'magenta', 'red', 'white' and 'yellow'
    coloredlogs.DEFAULT_FIELD_STYLES = {
        "asctime": {"color": "green"},
        "hostname": {"color": "magenta"},
        "levelname": {"bold": True, "color": "cyan"},
        "name": {"color": "white"},
        "programname": {"color": "cyan"},
        "username": {"color": "yellow"},
        "thread": {"color": "green"},
    }
    coloredlogs.DEFAULT_LEVEL_STYLES = {
        "critical": {"bold": True, "color": "red"},
        "debug": {"color": "green"},
        "error": {"color": "red"},
        "info": {},
        "notice": {"color": "magenta"},
        "spam": {"color": "green", "faint": True},
        "success": {"bold": True, "color": "green"},
        "verbose": {"color": "blue"},
        "warning": {"color": "yellow"},
    }
    coloredlogs.install()
    if trace_loggers:
        logging.Logger.set_pineboo_default_level(loglevel)

    logging.basicConfig(format=log_format, level=app_loglevel)
    # LOGGER.info("LOG LEVEL: %s", loglevel)
    disable_loggers = ["PyQt6.uic.uiparser", "PyQt6.uic.properties", "blib2to3.pgen2.driver"]
    for loggername in disable_loggers:
        modlogger = logging.get_logger(loggername)
        modlogger.setLevel(logging.WARN)

    for loggername in trace_loggers:
        modlogger = logging.get_logger(loggername)
        modlogger.setLevel(logging.TRACE)


def exec_main_with_profiler(options: "optparse.Values") -> int:
    """Enable profiler."""
    import cProfile
    import pstats
    import io
    from pstats import SortKey  # type: ignore

    profile = cProfile.Profile()
    profile.enable()
    ret = exec_main(options)
    profile.disable()
    string_io = io.StringIO()
    sortby = SortKey.TIME
    print_stats = pstats.Stats(profile, stream=string_io).sort_stats(sortby)
    print_stats.print_stats(40)
    print(string_io.getvalue())
    return ret


def init_cli(catch_ctrl_c: bool = True) -> None:
    """Initialize singletons, signal handling and exception handling."""

    def _excepthook(
        type_: Type["BaseException"], value: "BaseException", traceback: "TracebackType"
    ) -> None:
        import traceback as pytback

        pytback.print_exception(type_, value, traceback)

    # PyQt 5.5 o superior aborta la ejecución si una excepción en un slot()
    # no es capturada dentro de la misma; el programa falla con SegFault.
    # Aunque esto no debería ocurrir, y se debería prevenir lo máximo posible
    # es bastante incómodo y genera problemas graves para detectar el problema.
    # Agregamos sys.excepthook para controlar esto y hacer que PyQt6 no nos
    # dé un segfault, aunque el resultado no sea siempre correcto:
    sys.excepthook = _excepthook  # type: ignore[assignment]
    # -------------------
    if catch_ctrl_c:
        # Fix Control-C / KeyboardInterrupt for PyQt:
        import signal

        signal.signal(signal.SIGINT, signal.SIG_DFL)


# def init_gui() -> None:
#    """Create GUI singletons."""
#    from pineboolib.plugins.mainform.eneboo import eneboo
#    from pineboolib.plugins.mainform.eneboo_mdi import eneboo_mdi

#    eneboo.mainWindow = eneboo.MainForm()
#    eneboo_mdi.mainWindow = eneboo_mdi.MainForm()


def setup_gui(app: "QtWidgets.QApplication") -> None:
    """Configure GUI app."""

    from PyQt6 import QtGui

    noto_fonts = [
        "NotoSans-BoldItalic.ttf",
        "NotoSans-Bold.ttf",
        "NotoSans-Italic.ttf",
        "NotoSans-Regular.ttf",
    ]
    for fontfile in noto_fonts:
        QtGui.QFontDatabase.addApplicationFont(
            utils_base.filedir("./core/fonts/noto_sans", fontfile)
        )

    style_app: str = settings.CONFIG.value("application/style", "Fusion")
    app.setStyle(style_app)  # type: ignore

    default_font = settings.CONFIG.value("application/font", None)
    if default_font is None:
        font = QtGui.QFont("Noto Sans", 9)
        font.setBold(False)
        font.setItalic(False)
    else:
        # FIXME: FLSettings.readEntry does not return an array
        font = QtGui.QFont(
            default_font[0], int(default_font[1]), int(default_font[2]), default_font[3] == "true"
        )

    app.setFont(font)


def init_testing(file_name: str = "") -> None:
    """Initialize Pineboo for testing purposes."""
    settings.CONFIG.set_value("application/dbadmin_enabled", True)

    if application.PROJECT.dgi is not None:
        from pineboolib.application.database import pnconnectionmanager

        del application.PROJECT._conn_manager
        application.PROJECT._conn_manager = pnconnectionmanager.PNConnectionManager()

    else:
        qapp = call_qapplication(sys.argv + ["-platform", "offscreen"])

        init_logging(True)  # NOTE: Use pytest --log-level=0 for debug
        init_cli(catch_ctrl_c=False)

        LOGGER.info("PINEBOO TESTING %s.", application.PINEBOO_VER)

        application.PROJECT.set_app(qapp)

        dgi = dgi_module.load_dgi("qt", None)

        application.PROJECT.init_dgi(dgi)

    application.TESTING_MODE = True
    application.PROJECT.aq_app._inicializing = False
    conn = connection.connect_to_db(connection.IN_MEMORY_SQLITE_CONN)

    if not application.PROJECT.init_conn(connection=conn):
        raise Exception("No main connection was established. Aborting Pineboo load.")

    # Si hay un dump para cargar se carga y así agilizamos inicialización de módulos!
    if file_name:
        file_path = os.path.join(application.PROJECT.tmpdir, "cache", file_name)
        LOGGER.warning("Buscando fichero DUMP %s." % file_path)
        if os.path.exists(file_path):
            LOGGER.warning("Cargando DUMP.")
            drv = application.PROJECT.conn_manager.default().driver()
            db_api_con = drv.connection().connection
            db_api_cursor = db_api_con.cursor()
            with open(file_path, "r") as file_:
                db_api_cursor.executescript(file_.read())

    # application.PROJECT.no_python_cache = False
    _initialize_data()


def finish_testing(delete_tmpdir: bool = True) -> None:
    """Clear data from pineboo project."""

    from pineboolib.application import qsadictmodules
    from pineboolib.application.parsers.parser_mtd import pnormmodelsfactory

    application.PROJECT.conn_manager.manager().cleanupMetaData()
    application.PROJECT.actions = {}
    application.PROJECT.areas = {}
    application.PROJECT.modules = {}
    if application.PROJECT.main_window:
        application.PROJECT.main_window.initialized_mods_ = []

    qsadictmodules.QSADictModules.clean_all()
    application.PROJECT.conn_manager.finish()
    for item_name in list(pnormmodelsfactory.PROCESSED):  # Si es de sistema, no se elimina.
        if item_name.startswith("fl"):
            continue

        pnormmodelsfactory.PROCESSED.remove(item_name)

    if delete_tmpdir:
        LOGGER.warning("Deleting temp folder %s", application.PROJECT.tmpdir)
        try:
            shutil.rmtree(application.PROJECT.tmpdir)
        except Exception as error:
            LOGGER.warning(
                "No se ha podido borrar %s al limpiar cambios del test. %s",
                application.PROJECT.tmpdir,
                error,
            )

    if not os.path.exists(application.PROJECT.tmpdir):
        os.mkdir(application.PROJECT.tmpdir)


def exec_main(options: "optparse.Values") -> int:
    """
    Exec main program.

    Handles optionlist and help.
    Also initializes all the objects
    """

    if options.external_modules:
        LOGGER.warn("Using external modules from %s", options.external_modules)
        sys.path.insert(0, options.external_modules)

    init_cli()

    app_args: List[str] = sys.argv
    if not options.enable_gui:
        app_args += ["-platform", "offscreen"]

    application.ENABLE_ACLS = options.enable_acls
    application.USE_INTERACTIVE_GUI = options.enable_interactive_gui
    application.ENABLE_CALL_EXCEPTIONS = options.enable_call_exceptions
    application.PARSE_PROJECT_ON_INIT = options.parse_project_on_init
    if options.omit_no_python_tags:
        from pineboolib.application.parsers import parser_qsa

        parser_qsa.IGNORE_NO_PYTHON_TAGS = True

    application.PROJECT.set_app(call_qapplication(app_args))

    if options.enable_gui:
        setup_gui(application.PROJECT.app)

    if options.log_sql:
        application.LOG_SQL = True

    if options.flfiles_folder:
        application.PROJECT.USE_FLFILES_FOLDER = options.flfiles_folder
        if options.update_flfiles:
            application.UPDATE_FLFILES_FROM_FLFOLDER = True

    if options.trace_debug:
        # "sys.settrace" function could lead to arbitrary code execution
        sys.settrace(utils_base.traceit)  # noqa: DUO111

    if options.trace_signals:
        from pineboolib.loader.utils import monkey_patch_connect

        monkey_patch_connect()

    if options.enable_dbadmin:
        application.PROJECT.db_admin_mode = True
        settings.CONFIG.set_value("application/dbadmin_enabled", True)
    if options.enable_quick:
        application.PROJECT.db_admin_mode = False
        settings.CONFIG.set_value("application/dbadmin_enabled", False)

    if options.enable_preping:
        settings.CONFIG.set_value("application/preping", True)

    if options.main_form:
        settings.CONFIG.set_value("ebcomportamiento/main_form_name", options.main_form)

    application.PROJECT.load_version()

    if utils_base.is_deployed():
        logging.basicConfig(stream=sys.stdout, level=logging.INFO)

        utils_base.download_files()

    dgi = dgi_module.load_dgi(options.dgi, options.dgi_parameter)
    # if options.enable_gui:
    #    init_gui()

    if dgi.useDesktop() and not options.enable_gui and options.dgi != "qt":
        LOGGER.info(
            "Selected DGI <%s> is not compatible with <pineboo-core>. Use <pineboo> instead"
            % options.dgi
        )

    if not dgi.useDesktop() and options.enable_gui:
        LOGGER.info(
            "Selected DGI <%s> does not need graphical interface. Use <pineboo-core> for better results"
            % options.dgi
        )

    if not dgi.useMLDefault():
        # When a particular DGI doesn't want the standard init, we stop loading here
        # and let it take control of the remaining pieces.
        return dgi.alternativeMain(options)

    configdb = connection.config_dbconn(options)
    LOGGER.debug(configdb)
    application.PROJECT.init_dgi(dgi)

    lang = application.PROJECT.aq_app._multi_lang_id.lower()
    if lang == "c":
        lang = "es"
    application.PROJECT.aq_app.loadTranslationFromModule("sys", lang)

    if not configdb and dgi.useDesktop() and dgi.localDesktop():
        if not dgi.mobilePlatform():
            from pineboolib.loader.dlgconnect.conn_dialog import show_connection_dialog

            configdb = show_connection_dialog(application.PROJECT.app)
            if configdb is None:
                return 2
        else:
            application.PROJECT.db_admin_mode = True
            settings.CONFIG.set_value("application/dbadmin_enabled", True)
            configdb = connection.DEFAULT_SQLITE_CONN

    if not configdb:
        raise ValueError("No connection given. Nowhere to connect. Cannot start.")

    conn = connection.connect_to_db(configdb)

    if not application.PROJECT.init_conn(connection=conn):
        LOGGER.warning("No main connection was provided. Aborting Pineboo load.")
        return -99

    settings.SETTINGS.set_value("DBA/lastDB", conn.DBName())

    if options.no_python_cache:
        application.PROJECT.no_python_cache = options.no_python_cache

    if options.enable_gui:
        from pineboolib.plugins import mainform

        main_form_name = settings.CONFIG.value("ebcomportamiento/main_form_name", "eneboo")

        main_form = getattr(mainform, main_form_name, None)
        if main_form is None:
            settings.CONFIG.set_value("ebcomportamiento/main_form_name", "eneboo")
            raise Exception(
                "mainForm %s does not exits!!.Use 'pineboo --main_form eneboo' to restore default mainForm"
                % main_form_name
            )
        # else:
        #    main_form = getattr(main_form, main_form_name)

        application.PROJECT.main_window = main_form.MainForm()
    # main_form_ = getattr(application.PROJECT.main_form, "MainForm", None)
    application.PROJECT.message_manager().send("splash", "show")
    _initialize_data()

    # FIXME: move this code to pineboo.application
    application.PROJECT.message_manager().send(
        "splash", "showMessage", ["Cargando traducciones ..."]
    )
    application.PROJECT.aq_app.loadTranslations()

    from pineboolib.loader import init_project

    ret = init_project.init_project(
        dgi,
        options,
        application.PROJECT,
        application.PROJECT.main_window if dgi.useDesktop() else None,
        application.PROJECT.app,
    )
    return ret


def _initialize_data(is_framework: bool = False) -> None:
    """Initialize data."""

    from PyQt6 import QtCore

    application.ID_SESSION = QtCore.QDateTime.currentDateTime().toString(
        QtCore.Qt.DateFormat.ISODate
    )

    if is_framework:
        LOGGER.info("STARTUP_FRAMEWORK:(3/7) Loading database.")
    application.PROJECT.run()

    from pineboolib.application.acls import pnaccesscontrollists

    acl = pnaccesscontrollists.PNAccessControlLists()
    acl.init()

    if acl._access_control_list:
        if is_framework:
            LOGGER.info("STARTUP_FRAMEWORK:(4/7) Loading ACLS.")
        application.PROJECT.aq_app.set_acl(acl)

    # LOGGER.info("STARTUP_FRAMEWORK:(5/9) Loading area definitions.")
    # application.PROJECT.conn_manager.managerModules().loadIdAreas()
    # LOGGER.info("STARTUP_FRAMEWORK:(6/9) Loading module definitions.")
    # application.PROJECT.conn_manager.managerModules().loadAllIdModules()
    if is_framework:
        LOGGER.info("STARTUP_FRAMEWORK:(5/7) Loading modules. Making QSA Tree.")
    application.PROJECT.load_modules()
    if is_framework:
        LOGGER.info("STARTUP_FRAMEWORK:(6/7) Loading classes. Making QSA Tree.")
    application.PROJECT.load_classes()
    if is_framework:
        LOGGER.info("STARTUP_FRAMEWORK:(7/7) Loading orm models. Making QSA Tree. ")
    application.PROJECT.load_orm()
    if is_framework:
        LOGGER.info("STARTUP_FRAMEWORK: All processes completed. Continue ...")
    application.PROJECT.conn_manager.removeConn("default")
    application.PROJECT.conn_manager.removeConn("dbAux")


def call_qapplication(args: List[str] = []) -> "QtWidgets.QApplication":
    """Call to QApplication."""
    from PyQt6 import QtWidgets

    return QtWidgets.QApplication(args)



"""Test_flfieldDB module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.fllegacy import fldateedit
from datetime import datetime


class TestFLFieldDBString(unittest.TestCase):
    """TestFLFieldDBString Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test string FLFieldDB mode."""

        from pineboolib.application import qsadictmodules
        from pineboolib.application.database import pnsqlcursor
        from pineboolib.core.utils import utils_base
        from PyQt6 import QtWidgets  # type: ignore[import]

        cursor_1 = pnsqlcursor.PNSqlCursor("flmodules")
        cursor_1.select()
        cursor_1.setModeAccess(cursor_1.Insert)
        cursor_1.refreshBuffer()
        cursor_1.insertRecord(False)

        module_ = qsadictmodules.QSADictModules.from_project("formRecordflmodules")
        self.assertTrue(module_)

        cursor_2 = module_.cursor()
        field = module_.child("flfielddb_5")
        self.assertNotEqual(field, None)
        field.setValue("hola")

        self.assertEqual(cursor_1.valueBuffer("descripcion"), "hola")
        cursor_2.setValueBuffer("descripcion", "nueva hola.")
        self.assertEqual(field.value(), "nueva hola.")
        field.status()
        field.selectAll()
        field.setShowAlias(True)
        self.assertTrue(field.showAlias())
        self.assertTrue(field.showEditor())
        field.setKeepDisabled(False)

        comp_mode = field.autoCompletionMode()
        self.assertTrue(comp_mode)
        field.setAutoCompletionMode(comp_mode)
        field.refresh()
        field.refreshQuick()

        del field.editor_
        field.initFakeEditor()
        field.field_map_value_ = field
        field.setMapValue()

        field.setNoShowed()
        field.autoCompletionUpdateValue()
        field.searchValue()

        field_icono = module_.child("flfielddb_3")
        icono_file = utils_base.filedir("./core/images/icons", "flfielddb.png")
        field_icono.setPixmap(icono_file)
        pix = field_icono.pixmap()
        field_icono.setPixmapFromPixmap(pix)

        clb = QtWidgets.QApplication.clipboard()
        clb.setPixmap(pix)  # type: ignore [union-attr]
        field_icono.setPixmapFromClipboard()

        # module_.form.close()

    def test_button_in_empty_buffer(self) -> None:
        """Check that the button is displayed on a control that points to a non-existent field."""
        from pineboolib.fllegacy import flfielddb
        from pineboolib.application import qsadictmodules

        module_ = qsadictmodules.QSADictModules.from_project("formRecordflmodules")
        parent = module_.parent()
        new_field = flfielddb.FLFieldDB(parent)
        new_field.setObjectName("fake_control")
        new_field.setFieldName("tes_field")
        new_field.load()
        lay = parent.layout()
        lay.addWidget(new_field)

        field = module_.child("fake_control")
        self.assertTrue(field)
        field.showWidget()
        self.assertEqual(field._push_button_db.isHidden(), False)

    def test_fldateedit_empty_value(self) -> None:
        """Check if the empty value is 00-00-0000."""
        from pineboolib.fllegacy import flfielddb
        from pineboolib.application.metadata import pnfieldmetadata
        from pineboolib.application import qsadictmodules
        from pineboolib import application

        module_ = qsadictmodules.QSADictModules.from_project("formRecordflmodules")
        parent = module_.parent()
        table_mtd = application.PROJECT.conn_manager.manager().metadata("flmodules")
        field_mtd = pnfieldmetadata.PNFieldMetaData(
            "date_control",
            "Date",
            True,
            False,
            "date",
            0,
            False,
            True,
            False,
            0,
            0,
            False,
            False,
            False,
            None,
            False,
            None,
            True,
            False,
            False,
        )
        if table_mtd is not None:
            table_mtd.addFieldMD(field_mtd)
        new_field = flfielddb.FLFieldDB(parent)
        new_field.setName("date_control")
        self.assertEqual(new_field.objectName(), "date_control")
        new_field.setFieldName(field_mtd.name())
        new_field.load()
        cursor = new_field.cursor()
        self.assertEqual(module_.cursor(), cursor)
        field_mtd_2 = cursor.metadata().field("date_control")
        if field_mtd_2 is not None:
            self.assertEqual(field_mtd, field_mtd_2)
            self.assertEqual(field_mtd_2.type(), "date")
        editor = new_field.editor_
        if isinstance(editor, fldateedit.FLDateEdit):
            self.assertEqual(editor.DMY, "dd-MM-yyyy")
            editor.date = "01-02-2001"
            self.assertEqual(editor.date, "2001-02-01")
            editor.date = None
            self.assertEqual(editor.date, "")

            new_field.setValue("2011-03-02")
            self.assertEqual(str(new_field.value())[:10], "2011-03-02")

            new_field.refresh()
            new_field.refreshQuick()
            new_field.setActionName("nueva_action")
            self.assertEqual(new_field.actionName(), "nueva_action")

            new_field.setFilter("nuevo_filtro")
            self.assertEqual(new_field.filter(), "nuevo_filtro")

            new_field.setForeignField("foreignfield")
            self.assertEqual(new_field.foreignField(), "foreignfield")

            new_field.setFieldRelation("fieldrelation")
            self.assertEqual(new_field.fieldRelation(), "fieldrelation")

            new_field.toggleAutoCompletion()
            if table_mtd is not None:
                table_mtd.removeFieldMD(field_mtd.name())

                self.assertFalse(
                    application.PROJECT.conn_manager.useConn("default").regenTable(
                        "flmodules", table_mtd
                    )
                )

    def test_basic_2(self) -> None:
        """Test basics 2."""
        from pineboolib.application import qsadictmodules
        from pineboolib.fllegacy import fllineedit

        module_ = qsadictmodules.QSADictModules.from_project("formRecordflmodules")
        field = module_.child("flfielddb_2")
        self.assertTrue(field)
        self.assertEqual(field._text_label_db.text(), "Versión")
        field.setFieldAlias("Versión 2")
        self.assertEqual(field._text_label_db.text(), "Versión 2")

        self.assertTrue(
            isinstance(field.editor_, fllineedit.FLLineEdit),
            "El tipo de campo es %s y se espera qlineedit.QLineEdit" % type(field.editor_),
        )

        self.assertEqual(field.echoMode(), fllineedit.FLLineEdit.EchoMode.Normal)
        field.setEchoMode(fllineedit.FLLineEdit.EchoMode.Password)
        self.assertEqual(field.echoMode(), fllineedit.FLLineEdit.EchoMode.Password)

        self.assertTrue(field.cursor_)

    def test_basic_3(self) -> None:
        """Test basics 3."""
        from pineboolib.qsa import qsa

        cur_test = qsa.FLSqlCursor("fltest")
        cur_test.openFormInMode(cur_test.Insert, False)
        form_ = qsa.from_project("formRecordfltest").form
        self.assertEqual(form_.child("fdb_double").value(), 0.0)
        self.assertEqual(form_.child("fdb_double").editor_.text(), "0.0")
        self.assertEqual(form_.child("fdb_id").value(), None)
        self.assertEqual(form_.child("fdb_date").value(), "")
        self.assertEqual(form_.child("fdb_time").value(), "00:00:00")
        self.assertEqual(form_.child("fdb_uint").value(), 0)
        cursor = form_.cursor()

        self.assertFalse(form_.child("fdb_bool").value())
        cursor.setValueBuffer("bool_field", True)
        self.assertTrue(form_.child("fdb_bool").value())
        cursor.setValueBuffer("uint_field", 10)
        self.assertEqual(form_.child("fdb_uint").value(), 10)
        cursor.setValueBuffer("uint_field", 11.01)
        self.assertEqual(form_.child("fdb_uint").value(), 11)
        cursor.setValueBuffer("uint_field", 12.99)
        self.assertEqual(form_.child("fdb_uint").value(), 12)
        cursor.setValueBuffer("double_field", 6.99)
        self.assertEqual(form_.child("fdb_double").value(), 6.99)
        today = qsa.Date()
        time_now = datetime.now().time()
        cursor.setValueBuffer("date_field", today)
        self.assertEqual(str(form_.child("fdb_date").value())[:10], str(today)[:10])
        cursor.setValueBuffer("time_field", time_now)
        self.assertEqual(form_.child("fdb_time").value(), str(time_now)[:8])

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Test_flformrecorddb module."""
import unittest
from pineboolib.loader.main import init_testing, finish_testing


class TestFLFormDB(unittest.TestCase):
    """TestFLFormDB Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_close_again(self) -> None:
        """Test closed 2 times."""
        from pineboolib.fllegacy import flformdb
        from pineboolib import application

        action = application.PROJECT.conn_manager.manager().action("flareas")
        form = flformdb.FLFormDB(action, None)
        self.assertTrue(form)
        form.load()
        self.assertTrue(form._loaded)
        form.close()
        self.assertFalse(form._loaded)
        form.close()
        self.assertFalse(form._loaded)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""QCombobox module."""

# -*- coding: utf-8 -*-
from PyQt6 import QtWidgets  # type: ignore[import]


from typing import Optional, Any, List, Union


class QComboBox(QtWidgets.QComboBox):
    """QComboBox class."""

    def __init__(
        self, parent: Optional[QtWidgets.QWidget] = None, name: Optional[str] = None
    ) -> None:
        """Inicialize."""

        super().__init__(parent)
        if name is not None:
            self.setObjectName(name)

        self.setEditable(False)

    def insertStringList(self, strl: List[str]) -> None:
        """Set items from an string list."""

        self.insertItems(len(strl), strl)

    def setReadOnly(self, data: bool) -> None:
        """Set read only."""

        super().setEditable(not data)

    def getCurrentItem(self) -> Any:
        """Return current item selected."""

        return super().currentIndex

    def setCurrentItem(self, new_item: Union[str, int]) -> None:
        """Set current item."""

        pos = -1
        if isinstance(new_item, str):
            pos = 0
            size_ = self.model().rowCount()  # type: ignore [union-attr]
            for pos_item in range(size_):
                item = self.model().index(pos_item, 0)  # type: ignore [union-attr]
                if item.data() == new_item:
                    pos = pos_item
                    break

        else:
            pos = new_item

        super().setCurrentIndex(pos)

    def getCurrentText(self) -> str:
        """Return current item text."""

        return super().currentText()

    def setCurrentText(self, value: Optional[str]) -> None:
        """Set current item text."""

        super().setCurrentText(value)  # type: ignore [arg-type]

    def setSizeLimit(self, size: int) -> None:
        """Set size limit."""

        super().setMaxCount(size)

    def getSizeLimit(self) -> int:
        """Return size limit."""

        return super().maxCount()

    sizeLimit = property(getSizeLimit, setSizeLimit, None, "get/set size allowed items limits")
    currentItem = property(getCurrentItem, setCurrentItem, None, "get/set current item index")
    currentText: str = property(  # type: ignore [assignment] # noqa F821
        getCurrentText, setCurrentText, None, "get/set current text"
    )



"""Test utils module."""

from pineboolib import application
import unittest
import threading

from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.qsa import qsa, utils


class TestUtils(unittest.TestCase):
    """Test Utils module."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_switch(self) -> None:
        """Test switch function."""

        for i in range(4):
            result = None
            for case in qsa.switch(i):
                if case(0):
                    result = 0
                    break
                if case(1):
                    result = 1
                    break
                if case(2):
                    result = 2
                    break
                if case():
                    result = 4
            if i < 3:
                self.assertEqual(i, result)
            else:
                self.assertFalse(i == result)

    def test_math(self) -> None:
        """Test Math class."""
        math_ = qsa.Math
        self.assertEqual(math_.abs(-1), 1)
        self.assertEqual(math_.abs(-2), 2)
        self.assertEqual(math_.abs(0), 0)

        self.assertEqual(math_.ceil(8.001), 9)
        self.assertEqual(math_.ceil(8.51), 9)
        self.assertEqual(math_.ceil(8.99), 9)

        self.assertEqual(math_.floor(8.001), 8)
        self.assertEqual(math_.floor(8.51), 8)
        self.assertEqual(math_.floor(8.99), 8)

        self.assertEqual(math_.pow(2, 1), 2)
        self.assertEqual(math_.pow(3, 2), 9)

        self.assertEqual(math_.round(10.1234, 2), 10.12)
        self.assertEqual(math_.round(0.9698, 2), 0.97)
        self.assertEqual(math_.round(123.969899, 4), 123.9699)
        self.assertEqual(math_.round(123.969899), 124)

        self.assertTrue(math_.random() > 0)
        self.assertTrue(math_.max(1, 2) == 2)
        self.assertTrue(math_.min(50, 33) == 33)
        self.assertTrue(math_.min(50, 50.01) == 50)
        self.assertTrue(math_.sqrt(64) == 8)
        self.assertEqual(math_.tan(20), 2.237160944224742)
        self.assertEqual(math_.cos(20), 0.40808206181339196)
        self.assertEqual(math_.acos(0.4), 1.1592794807274085)

        self.assertEqual(math_.E, 2.718281828459045)
        self.assertEqual(math_.PI, 3.141592653589793)
        self.assertEqual(math_.LN2, 0.6931471805599453)
        self.assertEqual(math_.LN10, 2.302585092994046)
        self.assertEqual(math_.LOG2E, 1.44269504089)
        self.assertEqual(math_.LOG10E, 0.4342944819)
        self.assertEqual(math_.SQRT1_2, 0.7071067811865476)
        self.assertEqual(math_.SQRT2, 1.4142135623730951)

    def test_parse_int(self) -> None:
        """Test parse_int function."""

        val_1 = qsa.parseInt("123", 10)
        self.assertEqual(val_1, 123)
        val_2 = qsa.parseInt("11", 2)
        self.assertEqual(val_2, 3)
        val_3 = qsa.parseInt("123,99", 10)
        self.assertEqual(val_3, 123)
        val_4 = qsa.parseInt("0xFE", 16)
        self.assertEqual(val_4, 254)
        val_5 = qsa.parseInt(100.0023, 10)
        self.assertEqual(val_5, 100)
        val_6 = qsa.parseInt(100, 2)
        self.assertEqual(val_6, 4)
        val_7 = qsa.parseInt("99")
        self.assertEqual(val_7, 99)

    def test_parse_float(self) -> None:
        """Test parse_float function."""

        val_1 = qsa.parseFloat(100)
        self.assertEqual(val_1, 100.0)
        val_2 = qsa.parseFloat(100.01)
        self.assertEqual(val_2, 100.01)
        val_3 = qsa.parseFloat("66000")
        self.assertEqual(val_3, 66000.0)
        val_4 = qsa.parseFloat("66000.2122")
        self.assertEqual(val_4, 66000.2122)
        val_5 = qsa.parseFloat("12:00:00")
        self.assertEqual(val_5, 12)
        val_6 = qsa.parseFloat("12:59:00")
        self.assertTrue(val_6 > 12.98 and val_6 < 12.99)

    def test_parse_string(self) -> None:
        """Test parse_string function."""

        val_1 = qsa.parseString(100)
        self.assertEqual(val_1, "100")

    def test_length(self) -> None:
        """Test length."""

        from pineboolib.application.types import Array

        list_ = ["uno", "dos", "tres"]
        dict_ = {"uno": 1, "dos": 2, "tres": 3, "cuatro": 4}
        array_1 = Array([1, 2, 3, 4, 5])
        array_2 = Array({"uno": 1, "dos": 2})

        self.assertEqual(qsa.length(list_), 3)
        self.assertEqual(qsa.length(dict_), 4)
        self.assertEqual(qsa.length(array_1), 5)
        self.assertEqual(qsa.length(array_2), 2)

    def test_is_nan(self) -> None:
        """Test isNaN."""

        self.assertTrue(qsa.isNaN("hola"))
        self.assertTrue(qsa.isNaN("0ct"))
        self.assertFalse(qsa.isNaN("0"))
        self.assertFalse(qsa.isNaN(11.21))
        self.assertFalse(qsa.isNaN("16.01"))

    def test_regexp(self) -> None:
        """Test regexp."""
        regexp = qsa.RegExp("d")
        self.assertFalse(regexp.global_)
        regexp.global_ = True
        self.assertTrue(regexp.global_)
        self.assertTrue(regexp.search("dog"))
        self.assertEqual(regexp.replace("dog", "l"), "log")
        self.assertEqual(regexp.cap(0), "d")
        self.assertEqual(regexp.cap(1), None)

    def test_replace(self) -> None:
        """Test replace."""
        regexp = qsa.RegExp("l")
        name = "pablo lopez"
        replace = utils.replace(name, regexp, "L")
        self.assertEqual(replace, "pabLo lopez")
        regexp.global_ = True
        replace2 = utils.replace(name, regexp, "L")
        self.assertTrue(isinstance(replace2, str))
        self.assertEqual(replace2, "pabLo Lopez")

        replace3 = utils.replace(replace2, "o", "6")
        self.assertEqual(replace3, "pabL6 L6pez")

    def test_timers(self) -> None:
        """Test Timers."""

        self.my_fun()
        timer_1 = qsa.startTimer(1000, self.my_fun)
        timer_2 = qsa.startTimer(1000, self.my_fun)  # noqa: F841
        timer_3 = qsa.startTimer(1000, self.my_fun)  # noqa: F841
        self.assertEqual(len(utils.TIMERS), 3)
        qsa.killTimer(timer_1)
        self.assertEqual(len(utils.TIMERS), 2)
        qsa.killTimers()
        self.assertEqual(len(utils.TIMERS), 0)

    def test_session(self) -> None:
        """Test session utils."""

        session_ = qsa.thread_session_new()
        self.assertEqual(session_, application.PROJECT.conn_manager.useConn("default").session())
        self.assertTrue(qsa.is_valid_session(session_))

    def test_type(self) -> None:
        """Test typeof function."""

        self.assertEqual(qsa.typeof_("hola"), "string")
        self.assertEqual(qsa.typeof_(True), "boolean")
        self.assertEqual(qsa.typeof_(False), "boolean")
        self.assertEqual(qsa.typeof_(8), "number")
        self.assertEqual(qsa.typeof_(1.01), "number")
        self.assertEqual(qsa.typeof_(0), "number")
        self.assertEqual(qsa.typeof_(qsa.Array()), "object")
        self.assertEqual(qsa.typeof_([]), "object")
        self.assertEqual(qsa.typeof_(qsa), "unknown")

    def test_thread(self) -> None:
        """Test Thread."""

        self.assertEqual(qsa.thread(), threading.current_thread().ident)

    def test_super(self) -> None:
        """Test super function."""

        obj_ = Prueba2()
        self.assertNotEqual(obj_.__class__, "Prueba")
        self.assertTrue(qsa._super("Prueba", obj_))  # type: ignore [arg-type] # noqa: F821

    def test_user_id(self) -> None:
        """Test user ids."""

        self.assertEqual(qsa.user_id(), "test3")
        qsa.set_user_id("pululo")
        self.assertEqual(qsa.user_id(), "pululo")

    def my_fun(self) -> None:
        """ "Callable test function."""
        print("EY")

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


class Prueba(object):
    """Prueba class."""

    pass


class Prueba2(Prueba):
    """Prueba2 class."""

    pass



"""StatusHelpMsg module."""

from pineboolib.core.utils import logging
from pineboolib.core import settings
from pineboolib import application

LOGGER = logging.get_logger(__name__)


class StatusHelpMsg(object):
    """StatusHelpMsg class."""

    def send(self, text_: str) -> None:
        """Send a text."""

        try:
            application.PROJECT.aq_app.statusHelpMsg(text_)

            if settings.CONFIG.value("ebcomportamiento/parser_qsa_gui", False):
                application.PROJECT.aq_app.popupWarn(text_)
        except RuntimeError as error:
            LOGGER.warning(str(error))



"""Test_signatures module."""

import unittest
from pineboolib.qsa import qsa
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.application.signatures.tests import fixture_path
import os


class TestSignatures(unittest.TestCase):
    """TestSignatures class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_qr(self) -> None:
        """test pdf_qr."""

        test_file = fixture_path("test.pdf")
        dest_file = fixture_path("test_result.pdf")

        self.assertTrue(os.path.exists(test_file))
        if os.path.exists(dest_file):
            os.remove(dest_file)

        obj_ = qsa.PdfQr(test_file)
        obj_.set_size(5)
        obj_.set_extension("PNG")
        obj_.set_dpi()
        obj_.set_text("textoa", "textob")
        obj_.set_font("Arial", 8)
        obj_.set_position(100, 100)
        self.assertTrue(obj_.sign(True))
        self.assertTrue(obj_.save_file(dest_file))
        self.assertTrue(obj_.get_qr())

        os.remove(dest_file)

    def test_xml_digest(self) -> None:
        """Test xml_digest."""

        xml_unsigned_file = fixture_path("xml_sin.xml")
        xml_signed_file = fixture_path("xml_con.xml")
        if os.path.exists(xml_signed_file):
            os.remove(xml_signed_file)

        cert_file = fixture_path("cert.p12")

        obj_ = qsa.XmlDigest(xml_unsigned_file, cert_file)
        obj_.set_password("123456")
        obj_.set_policy(
            [
                "http://www.facturae.gob.es/politica_de_firma_formato_facturae/politica_de_firma_formato_facturae_v3_1.pdf",
                "Politica de Firma FacturaE v3.1",
            ]
        )
        obj_.set_algorithm("sha1")
        self.assertTrue(obj_.sign())
        self.assertTrue(obj_._policy is not None)
        self.assertTrue(obj_.signature_value())

    """     def test_pdf_digest(self) -> None:
        """ "Test pdf_digest." """

        pdf_unsigned_file = fixture_path("pdf_unsigned.pdf")
        pdf_signed_file = fixture_path("pdf_signed.pdf")
        if os.path.exists(pdf_signed_file):
            os.remove(pdf_signed_file)

        cert_file = fixture_path("cert.p12")

        obj_ = qsa.PdfDigest(pdf_unsigned_file, cert_file)
        obj_.set_password("123456")
        obj_.set_stamp("stamp test")
        self.assertTrue(obj_.sign(pdf_signed_file))
        self.assertTrue(os.path.exists(pdf_signed_file))

        self.assertTrue(os.path.exists(pdf_signed_file)) """

    def test_xml_digest_512(self) -> None:
        """Test xml_digest."""

        xml_unsigned_file = fixture_path("xml_sin.xml")
        xml_signed_file = fixture_path("xml_con.xml")
        if os.path.exists(xml_signed_file):
            os.remove(xml_signed_file)

        cert_file = fixture_path("cert.p12")

        obj_ = qsa.XmlDigest(xml_unsigned_file, cert_file)
        obj_.set_password("123456")
        obj_.set_policy(
            [
                "http://www.facturae.gob.es/politica_de_firma_formato_facturae/politica_de_firma_formato_facturae_v3_1.pdf",
                "Politica de Firma FacturaE v3.1",
            ]
        )
        obj_.set_algorithm("sha512")
        self.assertTrue(obj_.sign())
        self.assertTrue(obj_.signature_value())

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""

        finish_testing()



"""Qtable module."""

# -*- coding: utf-8 -*-
from typing import Optional, Any, List, Union, cast
from PyQt6 import QtWidgets, QtCore, QtGui  # type: ignore[import]
from pineboolib.core import decorators
from pineboolib.core.utils.utils_base import format_double
from pineboolib.q3widgets import qwidget


# FIXMEQT6 class Q3TableWidget(QtWidgets.QTableWidget, qwidget.QWidget):
class Q3TableWidget(QtWidgets.QTableWidget, qwidget.QWidget):
    """
    Remove problematic properties from PyQt6-Stubs that we need to redefine.
    """

    currentChanged: Any


class QTable(Q3TableWidget):
    """QTable class."""

    linea_actuall = None
    currentChanged = QtCore.pyqtSignal(
        int, int
    )  # need overload (in Qt5, this signal is dataChanged)
    doubleClicked = QtCore.pyqtSignal(int, int)  # type: ignore [assignment] # noqa: F821
    clicked = QtCore.pyqtSignal(int, int)  # type: ignore [assignment] # noqa: F821
    valueChanged = QtCore.pyqtSignal(int, int)
    read_only_cols: List[int]
    read_only_rows: List[int]
    cols_list: List[str]
    resize_policy: QtWidgets.QSizePolicy

    Default = 0
    Manual = 1
    AutoOne = 2
    AutoOneFit = 3
    sort_column_: int

    def __init__(
        self, parent: Optional["QtWidgets.QGroupBox"] = None, name: Optional[str] = None
    ) -> None:
        """Inicialize."""
        super(QTable, self).__init__(parent)
        if not parent:
            self.setParent(self.parentWidget())

        if name:
            self.setObjectName(name)

        self.cols_list = []
        self.linea_actuall = -1
        cast(
            QtCore.pyqtSignal, self.currentCellChanged
        ).connect(  # type: ignore [attr-defined] # noqa: F821
            self.currentChanged_
        )
        cast(
            QtCore.pyqtSignal, self.cellDoubleClicked
        ).connect(  # type: ignore [attr-defined] # noqa: F821
            self.doubleClicked_
        )
        cast(
            QtCore.pyqtSignal, self.cellClicked
        ).connect(  # type: ignore [attr-defined] # noqa: F821
            self.simpleClicked_
        )
        cast(
            QtCore.pyqtSignal, self.itemChanged
        ).connect(  # type: ignore [attr-defined] # noqa: F821
            self.valueChanged_
        )
        self.read_only_cols = []
        self.read_only_rows = []
        self.resize_policy = cast(QtWidgets.QSizePolicy, 0)  # Default
        self.sort_column_ = -1

    def currentChanged_(
        self, current_row: int, current_column: int, previous_row: int, previous_column: int
    ) -> None:
        """Emit current changed signal."""
        if current_row > -1 and current_column > -1:
            cast(
                QtCore.pyqtSignal, self.currentChanged
            ).emit(  # type: ignore [attr-defined] # noqa: F821
                current_row, current_column
            )

    def doubleClicked_(self, field, col) -> None:
        """Emit double clicked signal."""
        cast(
            QtCore.pyqtSignal, self.cellDoubleClicked
        ).emit(  # type: ignore [attr-defined] # noqa: F821
            field, col
        )

    def simpleClicked_(self, field, col) -> None:
        """Emit simple clicked signal."""
        cast(QtCore.pyqtSignal, self.cellClicked).emit(  # type: ignore [attr-defined] # noqa: F821
            field, col
        )

    @decorators.not_implemented_warn
    def setResizePolicy(self, pol: QtWidgets.QSizePolicy) -> None:
        """Set resize polizy."""
        self.resize_policy = pol

    def __getattr__(self, name: str) -> Any:
        """Return an attribute specified by name."""
        if name == "Multi":
            return self.MultiSelection
        elif name == "SpreadSheet":
            return 999
        else:
            return getattr(super(), name, None)

    def valueChanged_(self, item=None) -> None:
        """Emit valueChanged signal."""

        if item and self.text(item.row(), item.column()) != "":
            cast(
                QtCore.pyqtSignal, self.valueChanged
            ).emit(  # type: ignore [attr-defined] # noqa: F821
                item.row(), item.column()
            )

    def numRows(self) -> int:
        """Return num rows."""

        return self.rowCount()

    def numCols(self) -> int:
        """Return num cols."""
        return self.columnCount()

    def setCellAlignment(self, row: int, col: int, alig_: int) -> None:
        """Set cell alignment."""
        self.item(row, col).setTextAlignment(alig_)  # type: ignore [union-attr]

    def setNumCols(self, num: int) -> None:
        """Set num cols."""
        self.setColumnCount(num)
        self.setColumnLabels(",", ",".join(self.cols_list))

    def setNumRows(self, num: int) -> None:
        """Set num rows."""
        self.setRowCount(num)

    def setReadOnly(self, read_only: bool) -> None:
        """Set read only."""

        self.setEditTriggers(
            QtWidgets.QAbstractItemView.EditTrigger.NoEditTriggers
            if read_only
            else QtWidgets.QAbstractItemView.EditTrigger.AllEditTriggers
        )

    def selectionMode(self) -> "QtWidgets.QAbstractItemView.SelectionMode":
        """Return selection mode."""
        return super(QTable, self).selectionMode()

    def setFocusStyle(self, style: Union[str, int]) -> None:
        """Set focus style."""

        if isinstance(style, int):
            return
        else:
            self.setStyleSheet(style)

    def setColumnLabels(self, separador: str, lista: str) -> None:
        """Set column labels."""
        array_ = lista.split(separador)
        if array_:
            self.setHorizontalHeaderLabels(array_[0 : self.columnCount()])

    def setRowLabels(self, separator: str, lista: str) -> None:
        """Set row labels."""
        array_ = lista.split(separator)
        if array_:
            self.setVerticalHeaderLabels(array_[0 : self.rowCount()])

    def clear(self) -> None:
        """Clear values."""

        super().clear()
        for i in range(self.rowCount()):
            self.removeRow(i)
        self.setHorizontalHeaderLabels(self.cols_list)
        self.setRowCount(0)

    def setSelectionMode(self, mode: "QtWidgets.QAbstractItemView.SelectionMode") -> None:
        """Set selection mode."""
        if mode.value == 999:  # type: ignore [comparison-overlap]
            self.setAlternatingRowColors(True)
        else:
            super().setSelectionMode(mode)

    def setColumnStrechable(self, col: int, value: bool) -> None:
        """Set column strechable."""
        if value:
            self.horizontalHeader().setSectionResizeMode(  # type: ignore [union-attr]
                col, QtWidgets.QHeaderView.ResizeMode.Stretch
            )
        else:
            self.horizontalHeader().setSectionResizeMode(  # type: ignore [union-attr]
                col, QtWidgets.QHeaderView.ResizeMode.ResizeToContents
            )

    def setHeaderLabel(self, label: str) -> None:
        """Set header label."""

        self.cols_list.append(label)
        self.setColumnLabels(",", ",".join(self.cols_list))

    def insertRows(self, numero, number: int = 1) -> None:
        """Insert Rows."""
        for rep in range(number):
            self.insertRow(numero)

    def text(self, row: int, col: int) -> str:
        """Return text from a index."""
        if row is None:
            return
        item = self.item(row, col)
        return item.text() if item else ""

    def setText(self, row: int, col: int, value: Any) -> None:
        """Set text to a index."""

        prev_item = self.item(row, col)
        if prev_item:
            bg_color = prev_item.background()

        right = True if isinstance(value, (int, float)) else False

        if right:
            value = value if isinstance(value, int) else format_double(value, len("%s" % value), 2)

        item = QtWidgets.QTableWidgetItem(str(value))

        if right:
            item.setTextAlignment(
                QtCore.Qt.AlignmentFlag.AlignVCenter + QtCore.Qt.AlignmentFlag.AlignRight
            )

        self.setItem(row, col, item)

        if prev_item:
            self.setCellBackgroundColor(row, col, bg_color)

        new_item = self.item(row, col)

        if new_item is not None:
            if row in self.read_only_rows or col in self.read_only_cols:
                new_item.setFlags(
                    cast(
                        QtCore.Qt.ItemFlag,
                        QtCore.Qt.ItemFlag.ItemIsSelectable | QtCore.Qt.ItemFlag.ItemIsEnabled,
                    )
                )
            else:
                new_item.setFlags(
                    cast(
                        QtCore.Qt.ItemFlag,
                        QtCore.Qt.ItemFlag.ItemIsSelectable
                        | QtCore.Qt.ItemFlag.ItemIsEnabled
                        | QtCore.Qt.ItemFlag.ItemIsEditable,
                    )
                )

    def setCellWidget(self, row: int, col: int, obj: Optional["QtWidgets.QWidget"]) -> None:
        """Set cell widget."""

        super().setCellWidget(row, col, obj)  # type: ignore [arg-type]

        widget = self.cellWidget(row, col)
        if widget is not None:
            if row in self.read_only_rows or col in self.read_only_cols:
                widget.setEnabled(False)

    def adjustColumn(self, k: int) -> None:
        """Adjust a column specified by name."""

        self.horizontalHeader().setSectionResizeMode(  # type: ignore [union-attr]
            k, QtWidgets.QHeaderView.ResizeMode.ResizeToContents
        )

    def setRowReadOnly(self, row: int, value: bool) -> None:
        """Set row read only specified by a number."""

        if value:
            if row in self.read_only_rows:
                return
            else:
                self.read_only_rows.append(row)
        else:
            if row in self.read_only_rows:
                self.read_only_rows.remove(row)
            else:
                return  # Ya esta en False la row

        for col in range(self.columnCount()):
            item = self.item(row, col)
            if item:
                if value:
                    item.setFlags(
                        cast(
                            QtCore.Qt.ItemFlag,
                            QtCore.Qt.ItemFlag.ItemIsSelectable | QtCore.Qt.ItemFlag.ItemIsEnabled,
                        )
                    )
                else:
                    item.setFlags(
                        cast(
                            QtCore.Qt.ItemFlag,
                            QtCore.Qt.ItemFlag.ItemIsSelectable
                            | QtCore.Qt.ItemFlag.ItemIsEnabled
                            | QtCore.Qt.ItemFlag.ItemIsEditable,
                        )
                    )

    def setColumnReadOnly(self, col: int, value: bool) -> None:
        """Set column read only."""

        if value:
            if col in self.read_only_cols:
                return
            else:
                self.read_only_cols.append(col)
        else:
            if col in self.read_only_cols:
                self.read_only_cols.remove(col)
            else:
                return

        for row in range(self.rowCount()):
            item = self.item(row, col)
            if item:
                if value:
                    item.setFlags(
                        cast(
                            QtCore.Qt.ItemFlag,
                            QtCore.Qt.ItemFlag.ItemIsSelectable | QtCore.Qt.ItemFlag.ItemIsEnabled,
                        )
                    )
                else:
                    item.setFlags(
                        cast(
                            QtCore.Qt.ItemFlag,
                            QtCore.Qt.ItemFlag.ItemIsSelectable
                            | QtCore.Qt.ItemFlag.ItemIsEnabled
                            | QtCore.Qt.ItemFlag.ItemIsEditable,
                        )
                    )

    @decorators.not_implemented_warn
    def setLeftMargin(self, margin: int):
        """Set left margin."""
        pass

    def setCellBackgroundColor(self, row: int, col: int, color: "QtGui.QBrush") -> None:
        """Set cell backgroun color."""
        item = self.item(row, col)

        if item is not None and color:
            item.setBackground(color)

    def getSorting(self) -> int:
        """Return sorting column."""

        return self.sort_column_

    def setSorting(self, col: int) -> None:
        """Set sorting column."""

        if not super().isSortingEnabled():
            super().setSortingEnabled(True)
        super().sortByColumn(col, QtCore.Qt.SortOrder.AscendingOrder)
        self.sort_column_ = col

    sorting = property(getSorting, setSorting)

    def editCell(self, row: int, col: int) -> None:
        """Edit a cell."""
        self.editItem(self.item(row, col))
